"""House price indices."""

import copy
import logging
from abc import ABC, abstractmethod
from typing import Any

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
import pandas as pd
from statsmodels.tsa.exponential_smoothing.ets import ETSModel

from hpipy.period_table import PeriodTable
from hpipy.price_model import BaseHousePriceModel, HedonicModel, RepeatTransactionModel
from hpipy.trans_data import HedonicTransactionData, RepeatTransactionData, TransactionData
from hpipy.utils.stineman_interpolation import interpolate_stineman


class BaseHousePriceIndex(ABC):
    """Abstract base house price index class."""

    data: TransactionData
    model: BaseHousePriceModel
    name: pd.Series
    periods: pd.Series
    value: pd.Series
    index: Any
    imputed: np.ndarray
    smooth: Any
    volatility: pd.DataFrame
    volatility_smooth: pd.DataFrame
    revision: pd.DataFrame
    revision_smooth: pd.DataFrame

    def __init__(self, **kwargs: Any) -> None:
        """Initialize base house price index."""
        for key, value in kwargs.items():
            setattr(self, key, value)

    @staticmethod
    def coef_to_index(
        coef_df: pd.DataFrame,
        log_dep: bool,
        base_price: float = 1,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Convert coefficients to an index.

        Args:
            coef_df (pd.DataFrame): Coefficients.
            log_dep (bool): Log dependent variable.
            base_price (float, optional): Base price.
                Defaults to 1.

        Returns:
            tuple[np.ndarray, np.ndarray]: Index and imputed values.

        """
        coef_df.loc[coef_df["coefficient"].abs() < 1e-15, "coefficient"] = 0
        coef_df.loc[coef_df["coefficient"].eq(0) & (coef_df.index > 0), "coefficient"] = np.nan

        # Set up imputation identification vector.
        is_imputed = np.zeros(len(coef_df["coefficient"]))

        # Determine which index values need to be imputed.
        na_coef = coef_df["coefficient"].isnull()

        # If any, then work through imputation process.
        if na_coef[1:].sum() == len(na_coef[1:]):
            coef_df[:] = 0
        elif na_coef.sum() > 0:
            # Set all missing to imputed.
            is_imputed[na_coef] = 1

            min_valid_idx = np.where(~na_coef & coef_df["coefficient"] != 0)[0].min()

            # Fix cases where beginning is missing.
            if coef_df[na_coef].index.min() < min_valid_idx:
                logging.info("You are extrapolating beginning periods.")
                not_na = np.where((~na_coef) & (coef_df["coefficient"] != 0))[0]
                imp_to_0 = na_coef.index < np.min(not_na)
                coef_df.loc[imp_to_0, "coefficient"] = 0

            # Fix cases where end is missing.
            if len(coef_df["coefficient"]) in np.where(na_coef)[0]:
                logging.info("You are extrapolating ending periods.")
                not_na = np.where(~na_coef)[0]
                end_imp = np.arange(np.max(not_na), len(coef_df["coefficient"] + 1))
                end_coef = coef_df["coefficient"].ffill()
                coef_df.loc[end_imp.tolist(), "coefficient"] = end_coef[end_imp.tolist()]

            # coef_df["coefficient"] = coef_df["coefficient"].interpolate(method="linear")
            coef_df["coefficient"] = coef_df["coefficient"].transform(
                lambda x: pd.Series(interpolate_stineman(x.values)),
            )

            n_imp = len(na_coef[na_coef])
            logging.info(f"Total of {n_imp} period{'s' if n_imp > 1 else ''} imputed.")

        max_period = int(coef_df["time"].max())

        # Convert estimate to an index value.
        if log_dep:  # coefficients represent log-ratio: log(Pt / P0)
            estimate = np.exp(coef_df["coefficient"])
            index_value = (estimate * 100)[:max_period]
        else:  # coefficients represent price delta: Pt - P0
            estimate = (coef_df["coefficient"] + base_price) / base_price
            index_value = (estimate * 100)[:max_period]
        index = index_value

        return index, is_imputed

    @classmethod
    def from_model(
        cls,
        model: BaseHousePriceModel,
        trans_data: TransactionData | None = None,
        max_period: int | None = None,
        smooth: bool = False,
        **kwargs: Any,
    ) -> Self:
        """Create an index from a house price model.

        Args:
            model (BaseHousePriceModel): House price model object.
            trans_data (TransactionData | None, optional): Transaction data
                object.
                Defaults to None.
            max_period (int | None, optional): Maximum period for the index.
                Defaults to None.
            smooth (bool, optional): Smooth the index.
                Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Self: Index object.

        """
        if max_period is None:
            max_period = int(model.coefficients["time"].max())

        # Check max period.
        if not isinstance(max_period, int | np.integer):
            msg = "'max_period' argument must be numeric/integer."
            raise ValueError(msg)

        # Extract coefficients.
        coef_df = pd.DataFrame({"time": list(range(1, max_period + 1))})
        coef_df = coef_df.merge(model.coefficients, on="time", how="left")

        log_dep = False
        if model.params.get("log_dep"):
            log_dep = True

        coef_cols = ["time", "coefficient"]
        if len(coef_df.columns) > 2:
            partition_cols = [col for col in coef_df.columns if col not in coef_cols]
            unique_df = coef_df[partition_cols].drop_duplicates()
            index = np.empty(len(unique_df))
            is_imputed = np.empty(len(unique_df))
            for idx, row in unique_df.iterrows():
                coef_df_i = coef_df.merge(row.to_frame().T, on=partition_cols, how="inner")
                index_i, is_imputed_i = cls.coef_to_index(
                    coef_df_i,
                    log_dep,
                    base_price=model.base_price,
                )
                index_i_df = pd.DataFrame({"index": index_i})
                for col in partition_cols:
                    index_i_df[col] = row[col]
                index[idx] = index_i_df
                is_imputed[idx] = is_imputed_i
        else:
            index, is_imputed = cls.coef_to_index(coef_df, log_dep, base_price=model.base_price)

        # Convert to a class object.
        instance = cls(
            model=model,
            name=model.periods["name"][:max_period],
            periods=model.periods["period"][:max_period],
            value=index,
            imputed=is_imputed[:max_period],
        )

        if not isinstance(instance, BaseHousePriceIndex):
            msg = "Converting model results to index failed."
            raise Exception(msg)

        if smooth:
            smooth_order = kwargs.get("smooth_order", 3)

            instance = instance.smooth_index(order=smooth_order, in_place=True)
            if not hasattr(instance, "smooth"):
                msg = "Smoothing index failed."
                raise Exception(msg)

        if trans_data is not None:
            instance.data = trans_data

        assert instance.model is not None
        assert instance.name is not None
        assert instance.periods is not None
        assert instance.value is not None
        assert instance.imputed is not None

        return instance

    @staticmethod
    def get_data() -> type[TransactionData]:
        """Abstract transaction data fetcher method."""
        raise NotImplementedError

    @staticmethod
    def get_model() -> type[BaseHousePriceModel]:
        """Abstract model fetcher method."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _create_transactions(
        cls,
        trans_data: TransactionData,
        *args: Any,
        **kwargs: Any,
    ) -> TransactionData:
        """Abstract transaction data creation method.

        Args:
            trans_data (TransactionData): Transaction data.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            TransactionData: Transaction data.

        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _create_model(cls, *args: Any, **kwargs: Any) -> BaseHousePriceModel:
        """Abstract model creation method."""
        raise NotImplementedError

    @classmethod
    def create_index(
        cls,
        trans_data: TransactionData | PeriodTable | pd.DataFrame,
        prop_id: str | None = None,
        trans_id: str | None = None,
        price: str | None = None,
        seq_only: bool = True,
        max_period: int | None = None,
        smooth: bool = False,
        periodicity: str | None = None,
        min_date: str | None = None,
        max_date: str | None = None,
        adj_type: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create the index.

        Args:
            trans_data (TransactionData | PeriodTable | pd.DataFrame): Input
                transaction data.
            prop_id (str | None, optional): Property identifier.
                Defaults to None.
            trans_id (str | None, optional): Transaction identifier.
                Defaults to None.
            price (str | None, optional): Price column name.
                Defaults to None.
            seq_only (bool, optional): Sequential only.
                Defaults to True.
            max_period (int | None, optional): Maximum index period.
                Defaults to None.
            smooth (bool, optional): Smooth the index.
                Defaults to False.
            periodicity (str | None, optional): Periodicity of the index.
                Defaults to None.
            min_date (str | None, optional): Minimum date for the index.
                Defaults to None.
            max_date (str | None, optional): Maximum date for the index.
                Defaults to None.
            adj_type (str | None, optional): Adjustment type.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            Self: Index object.

        """
        # Check if trans_data has a trans_df object.
        if not isinstance(trans_data, TransactionData):
            if not isinstance(trans_data, PeriodTable):
                if "date" not in kwargs:
                    msg = (
                        "When supplying a raw dataframe to the 'trans_df' object, a valid 'date' "
                        "argument must be supplied."
                    )
                    raise ValueError(msg)

                # Create period table object.
                trans_data = PeriodTable(trans_data).create_period_table(
                    date=kwargs["date"],
                    periodicity=periodicity,
                    min_date=min_date,
                    max_date=max_date,
                    adj_type=adj_type,
                )

            if trans_id is None:
                msg = (
                    "When not supplying an 'hpidata' object, a 'trans_id' argument must be "
                    "supplied."
                )
                raise ValueError(msg)

            if prop_id is None:
                msg = (
                    "When not supplying an 'hpidata' object, a 'prop_id' argument must be "
                    "supplied."
                )
                raise ValueError(msg)

            if price is None:
                msg = (
                    "When not supplying an 'hpidata' object, a 'price' argument must be supplied."
                )
                raise ValueError(msg)

            # Create transaction data object.
            trans_data = cls._create_transactions(
                trans_data,
                prop_id=prop_id,
                trans_id=trans_id,
                price=price,
                date=kwargs["date"],
                periodicity=periodicity,
                seq_only=seq_only,
            )

        if not hasattr(trans_data, "trans_df"):
            msg = "Converting sales data to sales object failed."
            raise ValueError(msg)

        model = cls._create_model(trans_data, **kwargs)

        if not isinstance(model, BaseHousePriceModel):
            msg = "Estimating model failed."
            raise Exception(msg)

        # Convert to an index.
        return cls.from_model(
            model,
            trans_data=trans_data,
            max_period=max_period,
            smooth=smooth,
            **kwargs,
        )

    def create_series(
        self,
        train_period: int = 12,
        max_period: int | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a series from the index.

        Args:
            train_period (int, optional): Train period for the series.
                Defaults to 12.
            max_period (int | None, optional): Maximum period for the series.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            Self: Index object.

        """
        if self.model is None:
            raise ValueError
        if self.data is None:
            raise ValueError

        # Check training period.
        if not isinstance(train_period, int | float):
            msg = "'train_period' must be a single numeric value."
            raise ValueError(msg)

        train_period = int(train_period)

        # Check for alternate max period and its allowed value.
        if max_period is None or max_period > self.model.periods.period.max():
            max_period = self.model.periods.period.max()

        if max_period is None:
            raise ValueError

        # Ensure training period is no greater than specified max or index max.
        if train_period >= min(self.model.periods.period.max(), max_period):
            msg = (
                "'train_period' is greater than the length of the "
                "index and/or the 'max_period' argument."
            )
            raise ValueError(msg)

        # Trim by time.
        time_range = list(range(train_period + 1, max_period + 1))
        time_range += [time_range[-1] + 1]  # ensures the proper number of periods are forecasted

        # Generate row ids for the training data.
        row_ids = [self.data.create_forecast_periods(time, train=True) for time in time_range]
        has_empty_period = any(len(row_id) == 0 for row_id in row_ids)
        if has_empty_period:
            logging.info("Some periods have no data. Removing them from series estimation.")
            time_range = [
                t for t in time_range if len(self.data.create_forecast_periods(t, train=True)) > 0
            ]
            row_ids = [row_id for row_id in row_ids if len(row_id) > 0]

        # Run models, indices and combine into HPI objects.
        is_hpis = []
        for idx in range(len(time_range)):
            data = copy.deepcopy(self.data)
            model = copy.deepcopy(self.model)
            if isinstance(self.data, TransactionData):
                data.trans_df = data.trans_df.iloc[row_ids[idx]]
            model_i = self._create_model(data, **model.params, **kwargs)
            index_i = copy.deepcopy(
                self.from_model(
                    model_i,
                    trans_data=data,
                    max_period=time_range[idx] - 1,
                ),
            )
            is_hpis.append(index_i)

        self.hpis = is_hpis

        return self

    def smooth_index(
        self,
        order: list[int] | int = 3,
        in_place: bool = False,
    ) -> Self | pd.Series:
        """Smooth the index.

        Args:
            order (list[int] | int, optional): Smoothing order.
                Defaults to 3.
            in_place (bool, optional): Smooth in place or return new object.
                Defaults to False.

        Returns:
            Self | pd.Series: Smoothed index or new smoothed index object.

        """
        if not isinstance(order, list):
            order = [order]

        if self.value is None:
            raise ValueError

        # Check order.
        for idx, o_i in enumerate(order):
            if np.all(
                isinstance(o_i, int | float)
                and not np.isnan(o_i)
                and o_i > 0
                and o_i <= len(self.value) / 2,
            ):
                order[idx] = int(round(o_i, 0))
            else:
                msg = (
                    "'order' argument must be a positive integer or list of positive integers "
                    " less than half the length of the index."
                )
                raise ValueError(msg)

        # Create smoothed index (retain existing).
        s_index = self.value

        # Smooth with moving average (multiple orders can be done in sequence).
        for o_i in order:
            s_index = s_index.rolling(window=o_i, center=True).mean()

        # Deal with NAs (should be NAs on the tail ends of the smoothing).
        na_smooth = np.where(s_index.isnull())[0]

        # Fill in low end NAs with original index.
        na_low = na_smooth[na_smooth < len(s_index) / 2]
        s_index.iloc[na_low] = self.value.iloc[na_low]

        # Fill in high-end NAs with forecasted values (off of smoothed).
        na_high = na_smooth[na_smooth >= len(s_index) / 2]
        high_fc = (
            ETSModel(s_index[: na_high[0]], error="add", trend=None, seasonal=None)
            .fit(disp=False)
            .forecast(steps=len(na_high))
        )

        new_high = (high_fc.mean() + self.value.iloc[na_high]) / 2
        s_index[na_high] = new_high
        sm_index = pd.Series(s_index, name="indexsmooth")

        self.order = order

        # If returning in place.
        if in_place:
            self.smooth = sm_index
            return self

        # If just returning result of smoothing.
        return sm_index

    def smooth_series(self, order: int = 3) -> Self:
        """Apply smoothing to all indices.

        Args:
            order (int) Smoothing order.
                Defaults to 3.

        Returns:
            Self: Index object.

        """
        for hpi in self.hpis:
            hpi = hpi.smooth_index(order, in_place=True)
        return self


class RepeatTransactionIndex(BaseHousePriceIndex):
    """Repeat transaction house price index.

    Implements the repeat transaction methodology for constructing
    house price indices. It uses pairs of transactions for the same
    property to estimate price changes over time, controlling for
    property-specific characteristics that remain constant between sales.

    Args:
        data: TransactionData. The transaction data used to construct the
            index.
        model: BaseHousePriceModel. The underlying price model used to
            estimate the index.

    Attributes:
        name: pd.Series. The names of the time periods in the index.
        periods: pd.Series. The time periods covered by the index.
        value: pd.Series. The index values.
        index: Any. The index values in an alternate format.
        imputed: np.ndarray. Boolean array indicating which periods were
            imputed.
        smooth: Any. Smoothed version of the index values.
        volatility: pd.DataFrame. Volatility measures for the index.
        volatility_smooth: pd.DataFrame. Smoothed volatility measures.
        revision: pd.DataFrame. Revision measures for the index.
        revision_smooth: pd.DataFrame. Smoothed revision measures.

    Example:
        >>> import pandas as pd
        >>> from hpipy.price_index import RepeatTransactionIndex
        >>> # Create sample transaction data.
        >>> data = pd.DataFrame({
        ...     "property_id": [1, 1, 2],
        ...     "transaction_id": [1, 2, 3],
        ...     "price": [200000, 250000, 300000],
        ...     "date": pd.to_datetime(["2020-01", "2021-01", "2020-02"]),
        ... })
        >>> # Create index.
        >>> index = RepeatTransactionIndex.create_index(
        ...     data,
        ...     prop_id="property_id",
        ...     trans_id="transaction_id",
        ...     price="price",
        ...     date="date",
        ...     periodicity="Y",
        ...     min_date="2020-01",
        ...     max_date="2021-02",
        ...     estimator="robust",
        ...     log_dep=True,
        ... )

    """

    @staticmethod
    def get_data() -> type[TransactionData]:
        """Get repeat transaction data."""
        return RepeatTransactionData

    @staticmethod
    def get_model() -> type[BaseHousePriceModel]:
        """Get repeat transaction model."""
        return RepeatTransactionModel

    @classmethod
    def _create_transactions(
        cls,
        trans_data: pd.DataFrame,
        *args: Any,
        **kwargs: Any,
    ) -> TransactionData:
        """Create repeat transaction data.

        Args:
            trans_data (pd.DataFrame): The transaction data used to construct
                the index.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            TransactionData: The repeat transaction data.

        """
        return cls.get_data()(trans_data).create_transactions(*args, **kwargs)

    @classmethod
    def _create_model(
        cls,
        trans_data: TransactionData,
        estimator: str,
        log_dep: bool,
        **kwargs: Any,
    ) -> BaseHousePriceModel:
        """Create a repeat transaction house price model.

        Args:
            trans_data (TransactionData): The transaction data used to
                construct the index.
            estimator (str): The estimator to use.
            log_dep (bool): Whether to use a log-dependent variable.
            **kwargs: Additional keyword arguments.

        Returns:
            BaseHousePriceModel: The repeat transaction house price model.

        """
        return cls.get_model()(trans_data).fit(estimator=estimator, log_dep=log_dep, **kwargs)


class HedonicIndex(BaseHousePriceIndex):
    """Hedonic house price index.

    Implements the hedonic methodology for constructing house price
    indices. It uses property characteristics to control for quality
    differences between properties and estimates the pure price change
    over time.

    Args:
        data: TransactionData. The transaction data used to construct the
            index.
        model: BaseHousePriceModel. The underlying price model used to
            estimate the index.

    Attributes:
        name: pd.Series. The names of the time periods in the index.
        periods: pd.Series. The time periods covered by the index.
        value: pd.Series. The index values.
        index: Any. The index values in an alternate format.
        imputed: np.ndarray. Boolean array indicating which periods were
            imputed.
        smooth: Any. Smoothed version of the index values.
        volatility: pd.DataFrame. Volatility measures for the index.
        volatility_smooth: pd.DataFrame. Smoothed volatility measures.
        revision: pd.DataFrame. Revision measures for the index.
        revision_smooth: pd.DataFrame. Smoothed revision measures.

    Example:
        >>> import pandas as pd
        >>> from hpipy.price_index import HedonicIndex
        >>> # Create sample transaction data.
        >>> data = pd.DataFrame({
        ...     "property_id": [1, 2, 3],
        ...     "transaction_id": [1, 2, 3],
        ...     "price": [200000, 250000, 300000],
        ...     "date": pd.to_datetime(["2020-01", "2021-01", "2020-02"]),
        ...     "sqft": [1500, 1800, 2000],
        ...     "bedrooms": [3, 3, 4],
        ... })
        >>> # Create index.
        >>> index = HedonicIndex.create_index(
        ...     data,
        ...     prop_id="property_id",
        ...     trans_id="transaction_id",
        ...     price="price",
        ...     date="date",
        ...     periodicity="A",
        ...     min_date="2020-01",
        ...     max_date="2021-01",
        ...     dep_var="price",
        ...     ind_var=["sqft", "bedrooms"],
        ... )

    """

    @staticmethod
    def get_data() -> type[TransactionData]:
        """Get hedonic transaction data."""
        return HedonicTransactionData

    @staticmethod
    def get_model() -> type[BaseHousePriceModel]:
        """Get hedonic model."""
        return HedonicModel

    @classmethod
    def _create_transactions(
        cls,
        trans_data: pd.DataFrame,
        *args: Any,
        **kwargs: Any,
    ) -> TransactionData:
        """Create hedonic transaction data.

        Args:
            trans_data (pd.DataFrame): The transaction data used to construct
                the index.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            TransactionData: The hedonic transaction data.

        """
        return cls.get_data()(trans_data).create_transactions(*args, **kwargs)

    @classmethod
    def _create_model(
        cls,
        trans_data: TransactionData,
        dep_var: str,
        ind_var: str,
        **kwargs: Any,
    ) -> BaseHousePriceModel:
        """Create a hedonic house price model.

        Args:
            trans_data (TransactionData): The transaction data used to construct
                the index.
            dep_var (str): The dependent variable.
            ind_var (str): The independent variables.
            **kwargs: Additional keyword arguments.

        Returns:
            BaseHousePriceModel: The hedonic house price model.

        """
        if dep_var is not None and ind_var is not None:
            return cls.get_model()(trans_data).fit(dep_var=dep_var, ind_var=ind_var, **kwargs)
        msg = "A dependent (dep_var) and independent variables (ind_var) must be provided."
        raise ValueError(msg)



"""hpiPy: House Price Indices in Python."""

import os

try:
    from .version import __version__
except ImportError:
    version_txt = os.path.join(os.path.dirname(__file__), "version.txt")
    with open(version_txt) as f:
        __version__ = f.read().strip()

from . import datasets, extensions, price_index, utils

__all__ = ["datasets", "extensions", "price_index", "utils"]



"""Stineman interpolation functions."""

import numpy as np


def stineman_slope(x: np.ndarray, y: np.ndarray, scale: bool = False) -> np.ndarray:
    """Estimate the slope of an interpolating function with Stineman's method.

    This function estimates the slope of a function that passes through a set
    of points in xy-plane. The slopes are calculated based on the tangent of
    circles passing through every three consecutive points as proposed by [1].

    Args:
        x (np.ndarray): Array of x-coordinates of points.
        y (np.ndarray): Array of y-coordinates of points.
        scale (bool, optional): If True, x and y values are normalized before
            the slope calculation.
            Defaults to False.

    Returns:
        np.ndarray: Array of estimated slopes of the interpolant at (x, y).

    References:
        [1] Stineman, Russell W. "A Consistently Well-Behaved Method of
            Interpolation." Creative Computing 6.7 (1980): 54-57.

    """
    m = len(x)
    if m == 2:
        yp = np.repeat(np.diff(y) / np.diff(x), 2)
    else:
        if scale:
            sx = np.diff(np.array([np.min(x), np.max(x)]))
            sy = np.diff(np.array([np.min(y), np.max(y)]))
            if sy <= 0:
                sy = 1  # type: ignore
            x = (x + 1) / sx
            y = y / sy
        dx = np.diff(x)
        dy = np.diff(y)
        yp = np.repeat(np.nan, m)
        dx2dy2p = dx[1:] ** 2 + dy[1:] ** 2
        dx2dy2m = dx[:-1] ** 2 + dy[:-1] ** 2
        yp[1:-1] = (dy[:-1] * dx2dy2p + dy[1:] * dx2dy2m) / (dx[:-1] * dx2dy2p + dx[1:] * dx2dy2m)
        s = dy[0] / dx[0]
        if ((s >= 0) & (s >= yp[1])) | ((s <= 0) & (s <= yp[1])):
            yp[0] = 2 * s - yp[1]
        else:
            yp[0] = s + np.abs(s) * (s - yp[1]) / (np.abs(s) + np.abs(s - yp[1]))
        s = dy[-1] / dx[-1]
        if ((s >= 0) & (s >= yp[-2])) | ((s <= 0) & (s <= yp[-2])):
            yp[-1] = 2 * s - yp[-2]
        else:
            yp[-1] = s + np.abs(s) * (s - yp[-2]) / (np.abs(s) + np.abs(s - yp[-2]))
        if scale:
            yp = yp * sy / sx
    return yp


def parabola_slope(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Estimate the slope of an interpolating parabola.

    This function uses the slope of a parabola passing through points to
    estimate the slope of an interpolant at these points.

    Args:
        x (np.ndarray): Array of x-coordinates of points.
        y (np.ndarray): Array of y-coordinates of points.

    Returns:
        np.ndarray: Estimate of the slope of the interpolant at (x, y).

    """
    m = len(x)
    dx = np.diff(x)
    dy = np.diff(y)
    dydx = dy / dx
    if m == 2:
        yp = np.repeat(dydx, 2)
    else:
        yp = [  # type: ignore
            (dydx[0] * (2 * dx[0] + dx[1]) - dydx[1] * dx[0]) / (dx[0] + dx[1]),
            (dydx[:-1] * dx[-1] + dydx[-1] * dx[:-1]) / (dx[-1] + dx[:-1]),
            (dydx[-1] * (2 * dx[-1] + dx[-2]) - dydx[-2] * dx[-1]) / (dx[-1] + dx[-2]),
        ]
    return yp


def calculate_stineman_interpolant(
    x: np.ndarray,
    y: np.ndarray,
    xout: np.ndarray,
    yp: np.ndarray | None = None,
    method: str | None = None,
) -> np.ndarray:
    """Calculate Stineman interpolation.

    The function calculates the interpolated values based on the method of
    Stineman [1]. The user can specify the method to compute the slope at
    given points if the slope is not known.

    Args:
        x (np.ndarray): Array of x-coordinates of points.
        y (np.ndarray): Array of y-coordinates of points.
        xout (np.ndarray): x-coordinate where the interpolant is to be found.
        yp (np.ndarray, optional): Slopes of the interpolating function at x.
            Optional: only given if they are known, otherwise the argument is
            not used.
            Defaults to None.
        method (str, optional): Method for computing the slope at the given
            points if the slope is not known. With method="stineman",
            Stineman's original method based on an interpolating circle is
            used. Use method="scaledstineman" if scaling of x and y is to be
            carried out before Stineman's method is applied, and use
            method="parabola" to calculate the slopes from a parabola through
            every three points.
            Defaults to None.

    Returns:
        np.ndarray: Array with components 'x' and 'y' with the coordinates of
            the interpolant at the points specified by `xout`.

    References:
        [1] Stineman, Russell W. "A Consistently Well-Behaved Method of
            Interpolation." Creative Computing 6.7 (1980): 54-57.

    """
    if x is None or y is None or xout is None:
        msg = "Wrong number of input arguments: `x`, `y` and `xout` must be specified."
        raise ValueError(msg)
    if (
        not isinstance(x, np.ndarray)
        or not isinstance(y, np.ndarray)
        or not isinstance(xout, np.ndarray)
        or not issubclass(x.dtype.type, np.number)
        or not issubclass(y.dtype.type, np.number)
        or not issubclass(xout.dtype.type, np.number)
    ):
        msg = "`x`, `y` and `xout` must be numeric vectors."
        raise ValueError(msg)
    if len(x) < 2:
        msg = "`x` must have 2 or more elements."
        raise ValueError(msg)
    if len(x) != len(y):
        msg = "`x` must have the same number of elements as `y`."
        raise ValueError(msg)
    if np.any(np.isnan(x)) or np.any(np.isnan(y)) or np.any(np.isnan(xout)):
        msg = "NaNs in `x`, `y` or xout are not allowed."
        raise ValueError(msg)
    if yp is not None:
        if not isinstance(yp, np.ndarray) or not issubclass(yp.dtype.type, np.number):
            msg = "`yp` must be a numeric vector."
            raise ValueError(msg)
        if len(y) != len(yp):
            msg = "When specified, `yp` must have the same number of elements as `y`."
            raise ValueError(msg)
        if np.any(np.isnan(yp)):
            msg = "NaNs in `yp` are not allowed."
            raise ValueError(msg)
        if method is not None:
            msg = "Method should not be specified if `yp` is given."
            raise ValueError(msg)

    dx = np.diff(x)
    dy = np.diff(y)

    if np.any(dx <= 0):
        msg = "The values of `x` must strictly increasing."
        raise ValueError(msg)

    # Calculation of slopes if needed.
    if yp is None:
        if (method is None) or (method == "scaledstineman"):
            yp = stineman_slope(x, y, scale=True)
        elif method == "stineman":
            yp = stineman_slope(x, y, scale=False)
        elif method == "parabola":
            yp = parabola_slope(x, y)

    # Preparations.
    m = len(x)
    s = dy / dx
    k = len(xout)

    ix = np.searchsorted(x, xout, side="right") - 1

    # For edgepoints, allow extrapolation within a tiny range.
    epx = 5 * (np.finfo(float).eps) * np.diff(np.array([np.min(x), np.max(x)]))
    ix[((np.min(x) - epx) <= xout) & (xout <= np.min(x))] = 0
    ix[(np.max(x) <= xout) & (xout <= (np.max(x) + epx))] = m - 2
    idx = (ix >= 0) & (ix <= (m - 2))
    ix1 = ix[idx]
    ix2 = ix1 + 1

    # Computation of the interpolant.
    # Three cases: dyo1dyo2 ==, > and < 0.
    dxo1 = xout[idx] - x[ix1]
    dxo2 = xout[idx] - x[ix2]
    y0o = y[ix1] + s[ix1] * dxo1
    dyo1 = (yp[ix1] - s[ix1]) * dxo1  # type: ignore
    dyo2 = (yp[ix2] - s[ix1]) * dxo2  # type: ignore
    dyo1dyo2 = dyo1 * dyo2
    yo = y0o

    # Skipped for m=2, in which case linear interpolation is sufficient,
    # unless slopes are given.
    if (m > 2) or (yp is not None):
        id = dyo1dyo2 > 0
        yo[id] = y0o[id] + dyo1dyo2[id] / (dyo1[id] + dyo2[id])
        id = dyo1dyo2 < 0
        yo[id] = y0o[id] + dyo1dyo2[id] * (dxo1[id] + dxo2[id]) / (dyo1[id] - dyo2[id]) / (
            dx[ix1][id]
        )

    yout = np.repeat(np.nan, k)
    yout[idx] = yo

    return yout


def fill_missing(arr: np.ndarray, option: str = "locf") -> np.ndarray:
    """Fill missing values.

    Missing values can be replaced with the the most recent present value
    prior to it in the series, known as the Last Observation Carried Forward
    (LOCF), or by the most recent present next value starting from the end
    of the series, known as the Next Observation Carried Backward (NOCB).

    Args:
        arr (np.ndarray): Array with missing values.
        option (str, optional): Algorithm to be used.
            Accepts the following input:
                - "locf": Last Observation Carried Forward
                - "nocb": Next Observation Carried Backward
            Defaults to "locf".

    Returns:
        np.ndarray: Array with missing values filled.

    """
    if option == "nocb":
        arr = np.flip(arr)
    mask = np.isnan(arr)
    idx = np.where(~mask, np.arange(mask.shape[0]), 0)
    val = np.maximum.accumulate(idx, axis=0)
    out = arr[val]
    if option == "nocb":
        out = np.flip(out)
    return out


def interpolate_stineman(x: np.ndarray) -> np.ndarray:
    """Interpolate missing values in an array using Stineman's method.

    To avoid NaNs at the beginning and end of the array, any remaining NaNs
    are forward filled (to eliminate end NaNs) and then backward filled (to
    eliminate beginning NaNs).

    Args:
        x (np.ndarray): Array with possible missing values.

    Returns:
        np.ndarray: Array with missing values replaced.

    """
    x_isnan = np.isnan(x)
    if x_isnan.sum() > 0:
        # Perform Stineman interpolation.
        x = calculate_stineman_interpolant(np.where(~x_isnan)[0], x[~x_isnan], np.arange(len(x)))
    if np.any(np.isnan(x)):
        # Fill to avoid NaNs at the beginning and end of array.
        x = fill_missing(fill_missing(x), option="nocb")
    return x



"""Data loader functions."""

from collections.abc import Callable
from typing import NamedTuple

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset


class Batch(NamedTuple):
    """Batch object."""

    x: dict[str, torch.Tensor]
    y: torch.Tensor
    batch_size: int


def collate_fn(
    batch: list[tuple[np.ndarray, np.ndarray]],
    preprocess_fn: Callable[[pd.DataFrame], dict[str, np.ndarray]],
    columns: list[str],
) -> Batch:
    """Collate a list of batch inputs.

    Feature data is transformed from a list of NumPy arrays into a single
    dictionary of batch-length tensors keyed by the feature column names.

    Response data is transformed from a list of NumPy arrays into a tensor.

    Args:
        batch (list[tuple[np.ndarray, np.ndarray]]): Inputs.
        preprocess_fn (Callable[[pd.DataFrame], dict[str, np.ndarray]]):
            Preprocessing function.
        columns (list[str]): List of column names.

    Returns:
        X as dict[str, torch.Tensor], y as torch.Tensor.

    """
    x, y = zip(*batch, strict=False)
    x = preprocess_fn(pd.DataFrame(np.stack(x), columns=columns))
    x = {k: torch.from_numpy(v) for k, v in x.items()}
    y = torch.from_numpy(np.stack(y))
    return Batch(x=x, y=y, batch_size=len(y))


class TabularDataset(Dataset):
    """Tabular dataset implementation for PyTorch."""

    def __init__(
        self,
        X: pd.DataFrame | pd.Series | np.ndarray,
        y: pd.Series | np.ndarray | None = None,
    ) -> None:
        """Initialize the tabular dataset.

        Args:
            X (pd.DataFrame | pd.Series | np.ndarray): Input feature data.
            y (pd.Series | np.ndarray | None, optional): Input
                response data.
                Defaults to None.

        """
        self.X = X
        self.y = y
        if isinstance(self.X, (pd.DataFrame | pd.Series)):
            self.X = self.X.to_numpy()
        if isinstance(self.y, pd.Series):
            self.y = self.y.to_numpy()

    def __getitem__(self, idx: int) -> tuple[np.ndarray, np.ndarray] | np.ndarray:
        """Return indexed X and y data."""
        X = self.X[idx]
        if self.y is not None:
            y = self.y[idx]
            return X, y
        return X

    def __len__(self) -> int:
        """Return length of data."""
        return len(self.X)



import numpy as np
import pandas as pd
import pytest

from hpipy.period_table import PeriodTable


@pytest.mark.usefixtures("toy_dataset")
def test_period_table(toy_dataset: pd.DataFrame) -> None:
    """Test period table creation."""
    toy_dataset["trans_date"] = toy_dataset["sale_date"]

    pt_df = PeriodTable(toy_dataset).create_period_table("sale_date", periodicity="annual")
    assert len(pt_df.period_table) == 7
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[3] == "2013"

    pt_df = PeriodTable(toy_dataset).create_period_table("sale_date", periodicity="quarterly")
    assert len(pt_df.period_table) == 28
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[3] == "2010-4"

    pt_df = PeriodTable(toy_dataset).create_period_table("sale_date", periodicity="monthly")
    assert len(pt_df.period_table) == 84
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[3] == "2010-Apr"

    pt_df = PeriodTable(toy_dataset).create_period_table("sale_date", periodicity="weekly")
    assert len(pt_df.period_table) == 364
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[5] == "week: 2010-02-07 to 2010-02-13"

    pt_df = PeriodTable(toy_dataset).create_period_table("sale_date", periodicity="equalfreq")
    assert len(pt_df.period_table) == 84
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[83] == "equalfreq (30): 2016-10-29 to 2016-12-22"

    pt_df = PeriodTable(toy_dataset).create_period_table(
        "sale_date",
        periodicity="equalfreq",
        freq=15,
    )
    assert len(pt_df.period_table) == 169
    assert len(pt_df.period_table.columns) == 4
    assert (pt_df.period_table["end_date"][168] - pt_df.period_table["start_date"][168]).days > 15

    pt_df = PeriodTable(toy_dataset).create_period_table(
        "sale_date",
        periodicity="equalfreq",
        freq=15,
        start="last",
    )
    assert len(pt_df.period_table) == 169
    assert len(pt_df.period_table.columns) == 4
    assert (pt_df.period_table["end_date"][0] - pt_df.period_table["start_date"][0]).days > 15

    pt_df = PeriodTable(toy_dataset).create_period_table(
        "sale_date",
        periodicity="equalfreq",
        freq=15,
        start="last",
        first_date="2010-01-01",
    )
    assert len(pt_df.period_table) == 169
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["start_date"].iloc[0].strftime("%Y-%m-%d") == "2010-01-01"
    assert (pt_df.period_table["end_date"][0] - pt_df.period_table["start_date"][0]).days > 15

    pt_df = PeriodTable(toy_dataset).create_period_table(
        "sale_date",
        periodicity="equalfreq",
        freq=15,
        start="last",
        first_date="2010-01-11",
    )
    assert len(pt_df.period_table) == 169
    assert len(pt_df.period_table.columns) == 4
    assert (pt_df.period_table["end_date"][0] - pt_df.period_table["start_date"][0]).days > 15

    pt_df = PeriodTable(toy_dataset).create_period_table(
        "sale_date",
        periodicity="equalsample",
        nbr_periods=50,
    )
    assert len(pt_df.period_table) == 50
    assert len(pt_df.period_table.columns) == 4
    assert pt_df.period_table["name"].iloc[49] == "period 50"
    assert (
        pt_df.period_table["end_date"].iloc[49] - pt_df.period_table["start_date"].iloc[49]
    ).days == 48

    toy_dataset_gap = toy_dataset.iloc[np.arange(1000, 5001, step=1000) - 1]
    pt = PeriodTable(toy_dataset_gap)
    assert len(pt.create_period_table("sale_date", periodicity="weekly").period_table) == 153
    assert len(pt.create_period_table("sale_date", periodicity="monthly").period_table) == 36
    assert len(pt.create_period_table("sale_date", periodicity="quarterly").period_table) == 13
    assert len(pt.create_period_table("sale_date", periodicity="annual").period_table) == 4

    with pytest.raises(TypeError):
        PeriodTable(toy_dataset).create_period_table()


@pytest.mark.usefixtures("toy_dataset")
def test_date_to_period(toy_dataset: pd.DataFrame) -> None:
    """Test date to period conversion."""
    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="monthly",
    )

    assert "trans_period" in sales_df.trans_df.columns
    assert "trans_date" in sales_df.trans_df.columns
    assert len(sales_df.period_table) == 84
    assert len(sales_df.period_table.columns) == 4

    with pytest.raises(KeyError):
        PeriodTable(toy_dataset[["pinx"]]).create_period_table(date="sale_date")

    with pytest.raises(TypeError):
        PeriodTable(toy_dataset).create_period_table(date="sale_price")

    # Check `min_date`.
    for min_date in ["2010-01-03", "2010-01-01"]:
        sales_df = PeriodTable(toy_dataset).create_period_table(
            date="sale_date",
            periodicity="monthly",
            min_date=min_date,
        )
        assert sales_df.min_date == min_date

    # Check `min_date` clip.
    for min_date in ["2010-01-03", "2010-01-01"]:
        sales_df = PeriodTable(toy_dataset).create_period_table(
            date="sale_date",
            periodicity="monthly",
            min_date=min_date,
            adj_type="clip",
        )
        assert sales_df.min_date == min_date

    # Check `min_date`.
    for max_date, expected_date in [("2016-12-29", "2016-12-29"), ("2010-12-27", "2016-12-22")]:
        sales_df = PeriodTable(toy_dataset).create_period_table(
            date="sale_date",
            periodicity="monthly",
            max_date=max_date,
        )
        assert sales_df.max_date == expected_date

    # Check `max_date` clip.
    for max_date in ["2016-12-21", "2016-12-29"]:
        sales_df = PeriodTable(toy_dataset).create_period_table(
            date="sale_date",
            periodicity="monthly",
            max_date=max_date,
            adj_type="clip",
        )
        assert sales_df.max_date == max_date

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="annual",
    )
    assert sales_df.trans_df["trans_period"].iloc[0] == 2

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2009-12-31",
        periodicity="annual",
    )
    assert sales_df.trans_df["trans_period"].iloc[0] == 2

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2011-12-31",
        adj_type="clip",
        periodicity="annual",
    )
    assert sales_df.trans_df["trans_period"].iloc[0] == 3

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="monthly",
    )
    assert sales_df.trans_df["trans_period"].max() == 84

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2009-12-31",
        periodicity="monthly",
    )
    assert sales_df.trans_df["trans_period"].max() == 84

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2011-12-31",
        adj_type="clip",
        periodicity="monthly",
    )
    assert sales_df.trans_df["trans_period"].max() == 60

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="q",
    )
    assert sales_df.trans_df["trans_period"].max() == 28

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2009-12-31",
        periodicity="q",
    )
    assert sales_df.trans_df["trans_period"].max() == 28

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2011-12-31",
        adj_type="clip",
        periodicity="q",
    )
    assert sales_df.trans_df["trans_period"].max() == 20

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2012-01-01",
        adj_type="clip",
        periodicity="q",
    )
    assert sales_df.trans_df["trans_period"].max() == 20

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="w",
    )
    assert sales_df.trans_df["trans_period"].max() == 364

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2009-12-31",
        periodicity="w",
    )
    assert sales_df.trans_df["trans_period"].max() == 364

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2011-12-31",
        adj_type="clip",
        periodicity="Weekly",
    )
    assert sales_df.trans_df["trans_period"].max() == 260

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        min_date="2012-01-01",
        adj_type="clip",
        periodicity="Weekly",
    )
    assert sales_df.trans_df["trans_period"].max() == 260

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="ef",
        freq=12,
    )
    assert sales_df.trans_df["trans_period"].max() == 212

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="equalfreq",
        freq=12,
        start="last",
    )
    assert sales_df.trans_df["trans_period"].max() == 212

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="equalfreq",
        freq=12,
        start="first",
        first_date="2009-08-01",
    )
    assert sales_df.trans_df["trans_period"].max() == 225

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="equalfreq",
        freq=12,
        start="end",
        last_date="2019-08-01",
    )
    assert sales_df.trans_df["trans_period"].max() == 213

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="es",
        nbr_periods=35,
    )
    assert sales_df.trans_df["trans_period"].max() == 35

    sales_df = PeriodTable(toy_dataset).create_period_table(
        date="sale_date",
        periodicity="es",
        nbr_periods=2,
    )
    assert sales_df.trans_df["trans_period"].max() == 2

    with pytest.raises(TypeError):
        PeriodTable(toy_dataset).create_period_table(date="sale_price")



"""Metrics utilities."""

import copy
import logging
from typing import Any

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold
from statsmodels.tsa.exponential_smoothing.ets import ETSModel

from hpipy.price_index import BaseHousePriceIndex, HedonicIndex, RepeatTransactionIndex
from hpipy.price_model import BaseHousePriceModel, HedonicModel, RepeatTransactionModel
from hpipy.trans_data import HedonicTransactionData, RepeatTransactionData, TransactionData


def accuracy(
    hpi_obj: BaseHousePriceIndex,
    test_method: str = "insample",
    test_type: str = "rt",
    pred_df: TransactionData | pd.DataFrame | None = None,
    smooth: bool = False,
    in_place: bool = False,
    in_place_name: str = "accuracy",
    **kwargs: Any,
) -> BaseHousePriceIndex | pd.DataFrame:
    """Calculate the accuracy of an index.

    Args:
        hpi_obj (BaseHousePriceIndex): Index object.
        test_method (str, optional): Testing method.
            Defaults to "insample".
        test_type (str, optional): Testing type.
            Defaults to "rt".
        smooth (bool, optional): Smooth the index. If True, the revision is
            calculated based on the smoothed indices.
            Defaults to False.
        in_place (bool, optional): Return accuracy in-place.
            Defaults to False.

    Returns:
        Index object containing the accuracy or DataFrame.

    """
    # Check for class of hpi_obj.
    if not isinstance(hpi_obj, BaseHousePriceIndex):
        msg = "'hpi_obj' must be an index instance."
        raise ValueError(msg)

    # Check for allowed test_method.
    if test_method not in ["insample", "kfold"]:
        msg = "'test_method' must be one of 'insample' or 'kfold'."
        raise ValueError(msg)

    # Check for allowed test_method.
    if test_type not in ["rt", "hed"]:
        msg = "'test_type' must be one of 'rt', 'hed'."
        raise ValueError(msg)

    # Check agreement between test_type and hpi_obj
    if isinstance(pred_df, pd.DataFrame):
        df: pd.DataFrame = pred_df
    else:
        df = hpi_obj.data.trans_df
    index_type = hpi_obj.model

    # Check for smooth.
    if smooth and not hasattr(hpi_obj, "smooth"):
        msg = "'hpi_obj' has no smoothed index. Please add one or set 'smooth' to False."
        raise ValueError(msg)

    # Clip pred_df to size of index.
    if test_type == "rt" and df["period_2"].max() > hpi_obj.periods.max():
        logging.info(
            f"Trimming prediction date down to period {hpi_obj.periods.max()} and before.",
        )
        df = df[df["period_2"] <= hpi_obj.periods.max()]

    # In-sample.
    if test_method == "insample":
        index = hpi_obj.value
        if smooth:
            index = hpi_obj.smooth
        accr_obj = insample_error(pred_df=df, index_type=index_type, index=index)

    # K-fold.
    if test_method == "kfold":
        accr_obj = kfold_error(hpi_obj=hpi_obj, pred_df=df, smooth=smooth, **kwargs)

    if in_place:
        setattr(hpi_obj, in_place_name, accr_obj)
        return hpi_obj

    return accr_obj


def series_accuracy(
    series_obj: BaseHousePriceIndex,
    test_method: str = "insample",
    test_type: str = "rt",
    pred_df: TransactionData | pd.DataFrame | None = None,
    smooth: bool = False,
    summarize: bool = False,
    in_place: bool = False,
    in_place_name: str = "accuracy",
    **kwargs: Any,
) -> BaseHousePriceIndex | pd.DataFrame | list[pd.DataFrame]:
    """Calculate the accuracy of a series.

    Args:
        series_obj (BaseHousePriceIndex): Series object.
        test_method (str, optional): Testing method.
            Defaults to "insample".
        test_type (str, optional): Testing type.
            Defaults to "rt".
        smooth (bool, optional): Smooth the index. If True, the revision is
            calculated based on the smoothed indices.
            Defaults to False.
        summarize (bool, optional): Summarize the accuracy.
            Defaults to False.
        in_place (bool, optional): Return accuracy in-place.
            Defaults to False.
        in_place_name (str, optional): In-place attribute name.
            Defualts to "accuracy".

    Returns:
        Series object containing the accuracy or DataFrame.

    """
    # Check for allowed test_method.
    if test_method not in ["insample", "kfold", "forecast"]:
        msg = "'test_method' must be one of 'insample', 'kfold' or 'forecast'."
        raise ValueError(msg)

    # Check for allowed test_method.
    if test_type not in ["rt", "hed"]:
        msg = "'test_type' must be one of 'rt', 'hed'."
        raise ValueError(msg)

    # Check agreement between test_type and hpi_obj.
    if not (
        (isinstance(series_obj.data, RepeatTransactionData) and test_type == "rt")
        or (isinstance(series_obj.data, HedonicTransactionData) and test_type == "hed")
    ):
        if pred_df is None:
            msg = (
                f"When 'test_type' ({test_type}) does not match the 'hpi' object model type "
                f"({type(series_obj)}) you must provide an 'pred_df' object of the necessary "
                f"class, in this case: {test_type}."
            )
            raise ValueError(msg)
    else:
        trans_data: TransactionData = series_obj.data

    accr_dfs = []
    if test_method != "forecast":
        # Check for smooth indices.
        if smooth and not hasattr(series_obj.hpis[0], "smooth"):
            msg = "No smoothed indices found. Please add or set smooth to False."
            raise ValueError(msg)

        # Calculate accuracy.
        for _, hpi_obj in enumerate(series_obj.hpis):
            hpi_obj_i = copy.deepcopy(series_obj)
            hpi_obj_i.data = hpi_obj.data
            hpi_obj_i.value = hpi_obj.value
            if smooth:
                hpi_obj_i.smooth = hpi_obj.smooth

            accr_df_i = pd.DataFrame(
                accuracy(
                    hpi_obj_i,
                    test_method,
                    test_type,
                    trans_data,
                    smooth=smooth,
                    in_place=False,
                ),
            )
            accr_dfs.append(accr_df_i)

        accr_df = pd.concat(accr_dfs, ignore_index=True)

        # If summarizing.
        if summarize:
            accr_df = (
                accr_df.groupby("pair_id")
                .agg({"pred_price": "mean", "error": "mean", "log_error": "mean"})
                .reset_index()
            )

    # If it is a forecast method.
    else:
        accr_df = forecast_error(is_obj=series_obj, trans_data=trans_data, smooth=smooth, **kwargs)

    # Return if not in place.
    if not in_place:
        return accr_df

    # Add to series object.
    if smooth and in_place_name == "accuracy":
        in_place_name = "accuracy_smooth"
    setattr(series_obj, in_place_name, accr_df)

    return series_obj


def insample_error(
    pred_df: pd.DataFrame,
    index_type: BaseHousePriceModel,
    index: pd.Series,
) -> pd.DataFrame:
    """Calculate in-sample error."""
    if not isinstance(pred_df, pd.DataFrame):
        msg = "'pred_df' argument must be a dataframe."
        raise ValueError(msg)
    if isinstance(index_type, RepeatTransactionModel):
        return insample_error_rtdata(pred_df, index)
    if isinstance(index_type, HedonicModel):
        return insample_error_heddata(pred_df, index)
    raise ValueError


def insample_error_rtdata(pred_df: pd.DataFrame, index: pd.Series) -> pd.DataFrame:
    """Calculate in-sample error for repeat transaction index and data."""
    # Calculate the index adjustment to apply.
    adj = []
    for idx2, idx1 in zip(pred_df["period_2"] - 1, pred_df["period_1"] - 1, strict=False):
        adj.append(index.iloc[int(idx2)] / index.iloc[int(idx1)])

    # Calculate a prediction price.
    pred_price = pred_df["price_1"] * adj

    # Measure the error (difference from actual).
    error = (pred_price - pred_df["price_2"]) / pred_df["price_2"]
    logerror = np.log(pred_price) - np.log(pred_df["price_2"])

    return pd.DataFrame(
        {
            "pair_id": pred_df["pair_id"],
            "rt_price": pred_df["price_2"],
            "pred_price": pred_price,
            "error": error,
            "log_error": logerror,
            "pred_period": pred_df["period_2"],
        },
    )


def insample_error_heddata(pred_df: pd.DataFrame, index: pd.Series, **kwargs: Any) -> pd.DataFrame:
    """Calculate in-sample error for hedonic index and data."""
    # Future method.
    raise NotImplementedError


def kfold_error(
    hpi_obj: BaseHousePriceIndex,
    pred_df: pd.DataFrame,
    smooth: bool = False,
    k: int = 10,
    seed: int = 0,
    **kwargs: Any,
) -> pd.DataFrame:
    """Calculate k-fold error for an index.

    Args:
        hpi_obj (BaseHousePriceIndex): Index object.
        pred_df (pd.DataFrame): Prediction data.
        smooth (bool, optional): Smooth the index.
            Defaults to  False.
        k (int, optional): Number of folds.
            Defaults to 10.
        seed (int, optional): Seed.
            Defualts to 0.

    Returns:
        pd.DataFrame: k-fold error.

    """
    # Check hpi_obj.
    if not isinstance(hpi_obj, HedonicIndex | RepeatTransactionIndex):
        msg = "hpi_obj argument must be of class 'hpi'."
        raise ValueError(msg)

    # Check pred_df.
    if not isinstance(pred_df, pd.DataFrame):
        msg = "'pred_df' argument must be a DataFrame."
        raise ValueError(msg)

    # Check k.
    if not isinstance(k, int | float) or k < 2:
        msg = "Number of folds ('k' argument) must be a positive integer greater than 1."
        raise ValueError(msg)

    # Check seed.
    if not isinstance(seed, int | float) or seed < 0:
        msg = "'seed' must be a non-negative integer."
        raise ValueError(msg)

    errors = []

    kf = KFold(n_splits=k, shuffle=True, random_state=seed)
    for _, test_idx in kf.split(hpi_obj.data.trans_df):
        train_df, score_df = create_kfold_data(test_idx, hpi_obj.data, pred_df)

        train_data = copy.deepcopy(hpi_obj.data)
        train_data.trans_df = train_df
        score_data = copy.deepcopy(hpi_obj.data)
        score_data.trans_df = score_df

        k_model = hpi_obj._create_model(
            train_data,
            log_dep=hpi_obj.model.params["log_dep"],
            estimator=hpi_obj.model.params["estimator"],
            **kwargs,
        )

        k_index = hpi_obj.from_model(k_model, max_period=hpi_obj.data.trans_df["period_2"].max())

        # Deal with smoothing.
        if not smooth:
            index = k_index.value
        else:
            smooth_order = 3
            if "smooth_order" in kwargs:
                smooth_order = kwargs["smooth_order"]
            index = k_index.smooth_index(order=smooth_order)

        # Calc errors.
        k_error = insample_error(score_df, index_type=hpi_obj.model, index=index)
        errors.append(k_error)

    return pd.concat(errors)


def create_kfold_data(
    score_ids: np.ndarray,
    full_data: TransactionData,
    pred_df: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Create k-fold for transaction data."""
    if isinstance(full_data, RepeatTransactionData):
        return create_kfold_data_rtdata(score_ids, full_data, pred_df)
    raise NotImplementedError


def create_kfold_data_rtdata(
    score_ids: np.ndarray,
    full_data: TransactionData,
    pred_df: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Create k-fold for repeat transaction data."""
    train_df = full_data.trans_df.iloc[
        ~full_data.trans_df.reset_index(drop=True).index.isin(score_ids)
    ]
    score_df = match_kfold(train_df, pred_df, full_data)
    return train_df, score_df


def match_kfold(
    train_df: pd.DataFrame,
    pred_df: pd.DataFrame,
    full_data: TransactionData,
) -> pd.DataFrame:
    """Match k-fold for transaction data."""
    if isinstance(full_data, RepeatTransactionData):
        return match_kfold_rtdata(train_df, pred_df)
    raise NotImplementedError


def match_kfold_rtdata(train_df: pd.DataFrame, pred_df: pd.DataFrame) -> pd.DataFrame:
    """Match k-fold for repeat transaction data."""
    trans_pair = list(train_df["trans_id1"] + "_" + train_df["trans_id2"])
    return pred_df[~(pred_df["trans_id1"] + "_" + pred_df["trans_id2"]).isin(trans_pair)]


def match_kfold_heddata(train_df: pd.DataFrame, pred_df: pd.DataFrame) -> pd.DataFrame:
    """Match k-fold for hedonic transaction data."""
    # Choose every other one.
    x1 = pred_df.iloc[::2]  # even rows
    x2 = pred_df.iloc[1::2]  # odd rows
    # return pred_df[np.unique(zip(x1, x2))]
    return pred_df.loc[pd.MultiIndex.from_tuples(zip(x1, x2, strict=False)).unique()]


def forecast_error(
    is_obj: BaseHousePriceIndex,
    trans_data: TransactionData,
    forecast_length: int = 1,
    **kwargs: Any,
) -> pd.DataFrame:
    """Calculate forecast error.

    Args:
        is_obj (BaseHousePriceIndex): Index object.
        trans_data (TransactionData): Transaction data.
        forecast_length (int, optional): Forecast length.
            Defaults to 1.

    Returns:
        pd.DataFrame: Forecast error.

    """
    # Check classes.
    if not isinstance(is_obj, HedonicIndex | RepeatTransactionIndex):
        msg = "First argument must be an index instance."
        raise ValueError(msg)
    if not isinstance(trans_data, HedonicTransactionData | RepeatTransactionData):
        msg = "'trans_data' argument must be a transaction data instance."
        raise ValueError(msg)

    if is_obj.hpis is None:
        raise ValueError

    # Set start and end.
    start = int(is_obj.hpis[0].value.index[-1] + 1)
    end = int(is_obj.hpis[-1].value.index[-1] + 1)
    time_range = list(range(start, end + 1))

    # Get data.
    fc_preddata = [
        trans_data.create_forecast_periods(time_cut, forecast_length=forecast_length, train=False)
        for time_cut in time_range
    ]

    # Predict value.
    if "smooth" in kwargs and kwargs["smooth"] and (len(is_obj.hpis[0].smooth) > 0):
        index_name = "smooth"
    else:
        if kwargs.get("smooth"):
            msg = "No smoothed indices found. Create them try again."
            raise ValueError(msg)
        index_name = "value"

    fit_kwargs = {"error": "add", "trend": None, "seasonal": None}
    fc_data = [
        (x.smooth.values if index_name == "smooth" else x.value.values) for x in is_obj.hpis
    ]
    fc_forecasts = [
        np.r_[x, ETSModel(x, **fit_kwargs).fit(disp=False).forecast(steps=forecast_length)]
        for x in fc_data
    ]
    if isinstance(trans_data, TransactionData):
        df = trans_data.trans_df
        index_type = is_obj.model
    else:
        raise ValueError

    if df is None:
        raise ValueError

    # Iterate through forecasts and calculate errors.
    fc_error = []
    for preddata, forecasts in zip(fc_preddata, fc_forecasts, strict=False):
        error = insample_error(
            pred_df=df.iloc[preddata],
            index_type=index_type,
            index=pd.Series(forecasts, index=np.arange(1, len(forecasts) + 1)),
        )
        fc_error.append(error)

    return pd.concat(fc_error)

    # self.error_df = error_df
    # self.test_method = 'forecast'

    # If returning forecasts.
    # if return_forecasts:
    #    self.forecasts = fc_forecasts


def revision(
    series_obj: BaseHousePriceIndex,
    in_place: bool = False,
    smooth: bool = False,
) -> BaseHousePriceIndex | pd.DataFrame:
    """Calculate the revision of a series.

    This is done by calculating the difference between consecutive indexes
    (n to n+1) and then calculating the mean and median of the differences.
    This is done for each period and the results are returned as a DataFrame.

    Args:
        series_obj (BaseHousePriceIndex): Series object.
        in_place (bool, optional): If True, the revision is placed into the
            series object as a named list.
            Defaults to False.
        smooth (bool, optional): Smooth the index. If True, the revision is
            calculated based on the smoothed indices.
            Defaults to False.

    Returns:
         Revision and associated statistics as object or DataFrame.

    """
    # Check class.
    if not isinstance(series_obj, BaseHousePriceIndex):
        msg = "'series_obj' must be an index instance."
        raise ValueError(msg)

    if smooth and len(series_obj.hpis[0].smooth) > 0:
        # index_name = "smooth"
        indices = [hpi_obj.smooth for hpi_obj in series_obj.hpis]
    else:
        if smooth:
            msg = "No smoothed indices found. Create them with 'smooth_series' and try again."
            raise Exception(msg)
        # index_name = "value"
        indices = [hpi_obj.value for hpi_obj in series_obj.hpis]

    # Calculate the differences in the indices (n to n+1).
    index_diffs = [(indices[i][:-1] - indices[i - 1]) for i in range(1, len(indices))]

    # Extract differences into lists by period (essentially transposing list).
    period_diffs = [list(x) for x in zip(*index_diffs, strict=False)]

    # Convert to vector format in correct order.
    period_diffs = [list(reversed(x)) for x in period_diffs]

    # Calculate the mean and medians.
    period_means = [np.mean(x) for x in period_diffs]
    period_medians = [np.median(x) for x in period_diffs]

    # Package and return.
    rev_obj = pd.DataFrame(
        {
            "period": range(1, len(period_means) + 1),
            "mean": period_means,
            "median": period_medians,
        },
    )

    if in_place:
        if smooth:
            series_obj.revision_smooth = rev_obj
        else:
            series_obj.revision = rev_obj
        return series_obj

    return rev_obj


def volatility(
    index: BaseHousePriceIndex | pd.DataFrame,
    window: int = 3,
    in_place: bool = False,
    smooth: bool = False,
) -> BaseHousePriceIndex | pd.DataFrame:
    """Calculate index volatility.

    Args:
        index (BaseHousePriceIndex | pd.DataFrame): Index object.
        window (int, optional): Window for calculations.
            Defaults to 3.
        in_place (bool, optional): Return index in place.
            Defaults to False.
        smooth (bool, optional): Smooth the index.
            Defaults to False.

    Returns:
        Index object with volatility calculation or DataFrame.

    """
    # Save index_obj for future returning.
    index_obj = index

    # Strip from HPI or HPIIndex objects.
    if isinstance(index_obj, HedonicIndex | RepeatTransactionIndex | BaseHousePriceIndex):
        if not smooth:
            df = index_obj.value
        else:
            df = index_obj.smooth
            # Check to make sure a NULL wasn't given by smooth.
            if df is None:
                msg = "No smoothed index present. Please set 'smooth = False'."
                raise ValueError(msg)
    else:
        df = index_obj

    if df is None:
        raise ValueError

    # Check window.
    if (
        isinstance(window, int | float)
        and not np.isnan(window)
        and window > 0
        and window <= len(df) / 2
    ):
        window = round(window)
    else:
        msg = (
            "'window' argument must be a positive integer less than half the length of the index."
        )
        raise ValueError(msg)

    # Calculate changes.
    deltas = pd.Series((df[1:].to_numpy() - df[:-1].to_numpy()) / df[:-1].to_numpy())

    # Calculate mean rolling std.
    iv = deltas.rolling(window, center=True).std()

    # Create object.
    vol_obj = pd.DataFrame(
        {
            "roll": iv,
            "mean": iv.mean(),
            "median": iv.median(),
        },
        index=df.index[iv.first_valid_index() : iv.last_valid_index() + 1],
    )

    # self.vol_obj = vol_obj
    # self.orig = index
    # self.window = window

    # If returing in place.
    if in_place:
        if smooth:
            index_obj.volatility_smooth = vol_obj
        else:
            index_obj.volatility = vol_obj
        return index_obj

    # If just returning result of volatility calculation.
    return vol_obj


def series_volatility(
    series_obj: BaseHousePriceIndex,
    window: int = 3,
    smooth: bool = False,
) -> BaseHousePriceIndex:
    """Calculate volatility for a series of indices.

    Args:
        series_obj (BaseHousePriceIndex): Object with index series.
        window (int, optional): Window for calculations.
            Defaults to 3.
        smooth (bool, optional): Smooth the index.
            Defaults to False.

    Returns:
        Index object with series volatility calculation.

    """
    for hpi in series_obj.hpis:
        hpi = volatility(hpi, window, smooth=smooth, in_place=True)
    return series_obj



"""Configure unit tests."""

import pandas as pd
import pytest

from hpipy.datasets import load_ex_sales, load_seattle_sales


class MyTestDataset:
    """Test dataset objects."""

    def __init__(self) -> None:
        """Initialize toy dataset class."""

    @property
    def dataset(self) -> pd.DataFrame:
        """Dataset."""
        return load_ex_sales()


@pytest.fixture(scope="session")
def toy_dataset() -> pd.DataFrame:
    """Return toy dataset class instance."""
    return MyTestDataset().dataset


class SeattleDataset:
    """Test dataset objects."""

    def __init__(self) -> None:
        """Initialize Seattle dataset class."""

    @property
    def dataset(self) -> pd.DataFrame:
        """Dataset."""
        return load_seattle_sales()


@pytest.fixture(scope="session")
def seattle_dataset() -> pd.DataFrame:
    """Return Seattle dataset class instance."""
    return SeattleDataset().dataset



import warnings

import pandas as pd
import pytest

from hpipy.extensions import RandomForestIndex


@pytest.mark.usefixtures("seattle_dataset")
@pytest.mark.parametrize("estimator", ["pdp"])
@pytest.mark.parametrize("log_dep", [True, False])
def test_rf_create_trans(seattle_dataset: pd.DataFrame, estimator: str, log_dep: bool) -> None:
    """Test creation of random forest index."""
    rf_index = RandomForestIndex().create_index(
        seattle_dataset,
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["tot_sf", "beds", "baths"],
        estimator=estimator,
        log_dep=log_dep,
        periodicity="monthly",
    )
    assert rf_index.model.params["estimator"] == estimator


@pytest.mark.usefixtures("seattle_dataset")
@pytest.mark.parametrize("log_dep", [True, False])
@pytest.mark.parametrize("sim_ids", [None, [0, 1]])
@pytest.mark.parametrize("sim_count", [None, 10])
@pytest.mark.parametrize("sim_per", [None, 0.1])
def test_rf_create_trans_small(
    seattle_dataset: pd.DataFrame,
    log_dep: bool,
    sim_ids: list[int] | None,
    sim_count: int | None,
    sim_per: float | None,
) -> None:
    """Test creation of random forest index with small dataset."""
    rf_index = RandomForestIndex().create_index(
        seattle_dataset.iloc[:100],
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["tot_sf", "beds", "baths"],
        estimator="pdp",
        log_dep=log_dep,
        periodicity="monthly",
        sim_ids=sim_ids,
        sim_count=sim_count,
        sim_per=sim_per,
    )
    assert rf_index.model.params["estimator"] == "pdp"


@pytest.mark.usefixtures("seattle_dataset")
def test_rf_create_trans_bad(seattle_dataset: pd.DataFrame) -> None:
    """Test creation of random forest index with bad data."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        with pytest.raises(ValueError):
            _ = RandomForestIndex().create_index(
                seattle_dataset.assign(
                    sale_price=lambda x: x["sale_price"] - x["sale_price"].min() - 1,
                ),
                date="sale_date",
                price="sale_price",
                trans_id="sale_id",
                prop_id="pinx",
                dep_var="price",
                ind_var=["tot_sf", "beds", "baths"],
                estimator="pdp",
                log_dep=True,
                periodicity="monthly",
            )

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        with pytest.raises(ValueError):
            _ = RandomForestIndex().create_index(
                seattle_dataset.assign(
                    sale_price=lambda x: x["sale_price"] - x["sale_price"].min() - 1,
                ),
                date="sale_date",
                price="sale_price",
                trans_id="sale_id",
                prop_id="pinx",
                dep_var="price",
                ind_var=["tot_sf", "beds", "baths"],
                estimator="x",
                log_dep=True,
                periodicity="monthly",
            )



"""Data utilities."""

import numpy as np
import pandas as pd
import torch


def prepare_dataframe(
    X: pd.DataFrame,
    feature_dict: dict[str, list[str]],
) -> dict[str, np.ndarray]:
    """Prepare dataframe feature data.

    Transforms a DataFrame of features into a dictionary of NumPy feature
    arrays of the appropriate types.

    Args:
        df (pd.DataFrame): Input feature data.
        feature_dict (dict[str, list[str]]): Input feature dictionary.

    Returns:
        Dictionary of feature names mapped to NumPy arrays.

    """
    X_out = {}

    for key in feature_dict["nulls"]:
        X_out[key] = np.array(X[[key]]).astype(np.float32)

    for key in feature_dict["numerics"] + feature_dict["log_numerics"]:
        X_out[key] = np.array(X[[key]]).astype(np.float32)

    for key in feature_dict["categoricals"] + feature_dict["ordinals"]:
        X_out[key] = np.array(X[[key]]).astype(np.int32)

    return X_out


def prepare_tensor(X: torch.Tensor, feature_dict: dict[str, list[str]]) -> dict[str, torch.Tensor]:
    """Prepare tensor feature data.

    Transforms a Tensor of features into a dictionary of feature tensors of
    the appropriate types.

    Args:
        X (torch.Tensor): Input feature data.
        feature_dict (dict[str, list[str]]): Input feature dictionary.

    Returns:
        Dictionary of feature names mapped to Tensors.

    """
    X_out = {}

    f_idx = 0
    for d in feature_dict:
        for key in feature_dict[d]:
            if (
                key in feature_dict["nulls"]
                or key in feature_dict["numerics"]
                or key in feature_dict["log_numerics"]
                or key in feature_dict["categoricals"]
                or key in feature_dict["ordinals"]
            ):
                X_out[key] = X[:, f_idx]
            else:
                continue
            f_idx += 1

    return X_out



from typing import Any
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytest
import torch
from torch import nn

from hpipy.extensions import NeuralNetworkIndex
from hpipy.extensions.neural.neural_avm.data_preprocessors import GeospatialPreprocessor
from hpipy.extensions.neural.neural_avm.model_layers import MonotonicDense, OrdinalEmbedding
from hpipy.extensions.neural.neural_avm.utils.data import prepare_dataframe, prepare_tensor
from hpipy.extensions.neural.neural_avm.utils.model import mixup, quantile_loss


@pytest.mark.usefixtures("seattle_dataset")
@pytest.mark.parametrize("estimator", ["residual", "attributional"])
def test_nn_create_trans(seattle_dataset: pd.DataFrame, estimator: str) -> None:
    """Test creation of neural network index."""
    nn_index = NeuralNetworkIndex().create_index(
        seattle_dataset,
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["area", "baths", "beds", "latitude", "longitude", "tot_sf"],
        estimator=estimator,
        feature_dict={
            "numerics": ["latitude", "longitude"],
            "log_numerics": ["tot_sf"],
            "categoricals": ["area", "sale_date"],
            "ordinals": ["baths", "beds"],
            "hpi": ["sale_date"],
        },
        num_models=1,
        num_epochs=3,
        min_pred_epoch=1,
        verbose=True,
    )
    assert nn_index.model.params["estimator"] == estimator


@pytest.fixture
def device() -> torch.device:
    """Return default device for testing."""
    return torch.device("cpu")


class TestMonotonicDense:
    """Test suite for MonotonicDense layer."""

    @pytest.fixture
    def layer_params(self) -> dict[str, Any]:
        """Return default layer parameters for testing."""
        return {
            "in_features": 3,
            "out_features": 6,
            "activation": nn.ReLU(),
            "monotonicity_indicator": [1, -1, 0],
            "activation_weights": [0.4, 0.4, 0.2],
        }

    def test_initialization(self, layer_params: dict[str, Any], device: torch.device) -> None:
        """Test layer initialization with various parameters."""
        layer = MonotonicDense(**layer_params, device=device)
        assert layer.in_features == layer_params["in_features"]
        assert layer.out_features == layer_params["out_features"]
        assert isinstance(layer.activation, nn.ReLU)

        if layer.monotonicity_indicator is not None:
            assert torch.allclose(
                layer.monotonicity_indicator,
                torch.tensor(layer_params["monotonicity_indicator"], device=device),
            )

        # Test activation weights normalization.
        expected_weights = torch.tensor(layer_params["activation_weights"], device=device)
        expected_weights = expected_weights / expected_weights.sum()
        assert torch.allclose(layer.activation_weights, expected_weights)

    def test_invalid_activation_weights(
        self,
        layer_params: dict[str, Any],
        device: torch.device,
    ) -> None:
        """Test that invalid activation weights raise appropriate errors."""
        with pytest.raises(ValueError):
            invalid_params = layer_params.copy()
            invalid_params["activation_weights"] = [1, 1]  # wrong length
            MonotonicDense(**invalid_params, device=device)

        with pytest.raises(ValueError):
            invalid_params = layer_params.copy()
            invalid_params["activation_weights"] = [-1, 1, 1]  # negative values
            MonotonicDense(**invalid_params, device=device)

    def test_forward_pass(self, layer_params: dict[str, Any], device: torch.device) -> None:
        """Test forward pass with different monotonicity settings."""
        layer = MonotonicDense(**layer_params, device=device)
        x = torch.randn(4, layer_params["in_features"], device=device)
        output = layer(x)

        assert output.shape == (4, layer_params["out_features"])
        assert not torch.isnan(output).any()

    def test_monotonicity_constraints(self, device: torch.device) -> None:
        """Test that monotonicity constraints are properly enforced."""
        # Check increasing monotonicity.
        layer = MonotonicDense(
            in_features=1,
            out_features=1,
            monotonicity_indicator=1,
            device=device,
        )
        x1 = torch.tensor([[1.0]], device=device)
        x2 = torch.tensor([[2.0]], device=device)
        y1 = layer(x1)
        y2 = layer(x2)
        assert (y2 >= y1).all()

        # Check decreasing monotonicity.
        layer = MonotonicDense(
            in_features=1,
            out_features=1,
            monotonicity_indicator=-1,
            device=device,
        )
        y1 = layer(x1)
        y2 = layer(x2)
        assert (y2 <= y1).all()


class TestOrdinalEmbedding:
    """Test suite for OrdinalEmbedding layer."""

    @pytest.fixture
    def layer_params(self) -> dict[str, Any]:
        """Return default layer parameters for testing."""
        return {
            "num_embeddings": 5,
            "embedding_dim": 3,
        }

    def test_initialization(self, layer_params: dict[str, Any], device: torch.device) -> None:
        """Test layer initialization."""
        layer = OrdinalEmbedding(**layer_params, device=device)
        assert layer.num_embeddings == layer_params["num_embeddings"]
        assert layer.embedding_dim == layer_params["embedding_dim"]
        assert layer.weight.shape == (
            layer_params["embedding_dim"],
            layer_params["num_embeddings"],
        )
        assert layer.null_weight.shape == (layer_params["embedding_dim"],)

    def test_forward_pass(self, layer_params: dict[str, Any], device: torch.device) -> None:
        """Test forward pass with various inputs."""
        layer = OrdinalEmbedding(**layer_params, device=device)

        # Test regular values.
        x = torch.tensor([1, 3, 2, 0, 4], device=device)
        output = layer(x)
        assert output.shape == (5, layer_params["embedding_dim"])
        assert not torch.isnan(output).any()

        # Test null values (zeros).
        x_null = torch.zeros(3, device=device)
        output_null = layer(x_null)
        assert output_null.shape == (3, layer_params["embedding_dim"])
        assert torch.allclose(output_null[0], output_null[1])

    def test_ordinal_encoding(self, device: torch.device) -> None:
        """Test that ordinal encoding is correct."""
        x = torch.tensor([1, 3, 2, 0, 4], device=device)

        expected_encoding = torch.tensor(
            [
                [1, 0, 0, 0],
                [1, 1, 1, 0],
                [1, 1, 0, 0],
                [0, 0, 0, 0],
                [1, 1, 1, 1],
            ],
            device=device,
        )

        with torch.no_grad():
            x = x.reshape(-1, 1)
            actual_encoding = (x > torch.arange(4, device=device)).float()

        assert torch.allclose(actual_encoding, expected_encoding.float())

    def test_parameter_shapes(self, layer_params: dict[str, Any], device: torch.device) -> None:
        """Test that parameter shapes are correct after initialization."""
        layer = OrdinalEmbedding(**layer_params, device=device)

        assert layer.weight.shape == (
            layer_params["embedding_dim"],
            layer_params["num_embeddings"],
        )
        assert layer.null_weight.shape == (layer_params["embedding_dim"],)
        assert layer.bias is None


@pytest.mark.parametrize("batch_size", [4, 8])
@pytest.mark.parametrize("feature_dim", [2, 10])
def test_mixup_shape(batch_size: int, feature_dim: int) -> None:
    """Test mixup output shapes."""
    X = torch.randn(batch_size, feature_dim)
    y = torch.randn(batch_size)

    mixed_X, y_a, y_b, lam = mixup(X, y, alpha=1.0)

    assert mixed_X.shape == X.shape
    assert y_a.shape == y.shape
    assert y_b.shape == y.shape
    assert lam.shape == (batch_size, 1)


def test_mixup_lambda_range() -> None:
    """Test mixup lambda values are in [0, 1]."""
    batch_size, feature_dim = 16, 4
    X = torch.randn(batch_size, feature_dim)
    y = torch.randn(batch_size)

    _, _, _, lam = mixup(X, y, alpha=1.0)

    assert torch.all(lam >= 0)
    assert torch.all(lam <= 1)


def test_mixup_deterministic() -> None:
    """Test mixup with fixed seed for reproducibility."""
    batch_size, feature_dim = 4, 2
    X = torch.randn(batch_size, feature_dim)
    y = torch.randn(batch_size)

    torch.manual_seed(0)
    result1 = mixup(X.clone(), y.clone(), alpha=1.0)

    torch.manual_seed(0)
    result2 = mixup(X.clone(), y.clone(), alpha=1.0)

    for r1, r2 in zip(result1, result2, strict=False):
        assert torch.allclose(r1, r2)


@pytest.mark.parametrize("batch_size", [4, 8])
def test_quantile_loss_shape(batch_size: int) -> None:
    """Test quantile loss output shape."""
    y_true = torch.randn(batch_size)
    y_pred = torch.randn(batch_size)
    quantile = torch.tensor(0.5)

    loss = quantile_loss(y_true, y_pred, quantile)
    assert loss.shape == y_true.shape


@pytest.mark.parametrize("quantile", [0.1, 0.5, 0.9])
def test_quantile_loss_values(quantile: float) -> None:
    """Test quantile loss computation for different quantiles."""
    y_true = torch.tensor([1.0, 2.0, 3.0])
    y_pred = torch.tensor([1.5, 1.5, 1.5])
    q = torch.tensor(quantile)

    loss = quantile_loss(y_true, y_pred, q)

    assert torch.all(loss >= 0)


def test_quantile_loss_zero() -> None:
    """Test quantile loss is zero when prediction equals true value."""
    y_true = torch.tensor([1.0, 2.0, 3.0])
    y_pred = torch.tensor([1.0, 2.0, 3.0])
    quantile = torch.tensor(0.5)

    loss = quantile_loss(y_true, y_pred, quantile)
    assert torch.all(loss == 0)


def test_prepare_dataframe() -> None:
    """Test prepare_dataframe function."""
    test_data = pd.DataFrame(
        {
            "null_feature": [1.0, np.nan, 3.0],
            "numeric_feature": [1.5, 2.5, 3.5],
            "log_numeric_feature": [10.0, 20.0, 30.0],
            "categorical_feature": [0, 1, 2],
            "ordinal_feature": [1, 2, 3],
        },
    )

    feature_dict = {
        "nulls": ["null_feature"],
        "numerics": ["numeric_feature"],
        "log_numerics": ["log_numeric_feature"],
        "categoricals": ["categorical_feature"],
        "ordinals": ["ordinal_feature"],
    }

    result = prepare_dataframe(test_data, feature_dict)

    assert isinstance(result, dict)
    assert len(result) == 5

    # Check data types.
    assert result["null_feature"].dtype == np.float32
    assert result["numeric_feature"].dtype == np.float32
    assert result["log_numeric_feature"].dtype == np.float32
    assert result["categorical_feature"].dtype == np.int32
    assert result["ordinal_feature"].dtype == np.int32

    # Check values.
    np.testing.assert_array_equal(
        result["numeric_feature"],
        np.array([[1.5], [2.5], [3.5]], dtype=np.float32),
    )
    np.testing.assert_array_equal(
        result["categorical_feature"],
        np.array([[0], [1], [2]], dtype=np.int32),
    )


def test_prepare_tensor() -> None:
    """Test prepare_tensor function."""
    test_tensor = torch.tensor(
        [
            [1.0, 1.5, 10.0, 0.0, 1.0],
            [float("nan"), 2.5, 20.0, 1.0, 2.0],
            [3.0, 3.5, 30.0, 2.0, 3.0],
        ],
        dtype=torch.float32,
    )

    feature_dict = {
        "nulls": ["null_feature"],
        "numerics": ["numeric_feature"],
        "log_numerics": ["log_numeric_feature"],
        "categoricals": ["categorical_feature"],
        "ordinals": ["ordinal_feature"],
    }

    result = prepare_tensor(test_tensor, feature_dict)

    assert isinstance(result, dict)
    assert len(result) == 5

    # Check that all outputs are tensors.
    for key in result:
        assert isinstance(result[key], torch.Tensor)

    # Check values for a few features.
    assert torch.equal(result["numeric_feature"], test_tensor[:, 1])
    assert torch.equal(result["categorical_feature"], test_tensor[:, 3])


@pytest.mark.usefixtures("seattle_dataset")
def test_nn_invalid_estimator(seattle_dataset: pd.DataFrame) -> None:
    """Test handling of invalid estimator type."""
    nn_index = NeuralNetworkIndex().create_index(
        seattle_dataset,
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["area", "baths", "beds"],
        estimator="invalid_estimator",
        num_models=1,
        num_epochs=1,
        min_pred_epoch=1,
    )
    assert nn_index.model.params["estimator"] == "residual"


@pytest.mark.usefixtures("seattle_dataset")
def test_nn_no_log_dep(seattle_dataset: pd.DataFrame) -> None:
    """Test handling of non-log dependent variable."""
    with pytest.raises(NotImplementedError):
        NeuralNetworkIndex().create_index(
            seattle_dataset,
            date="sale_date",
            price="sale_price",
            trans_id="sale_id",
            prop_id="pinx",
            dep_var="price",
            ind_var=["area", "baths", "beds"],
            log_dep=False,
            num_models=1,
            num_epochs=1,
            min_pred_epoch=1,
        )


@pytest.mark.usefixtures("seattle_dataset")
def test_nn_missing_geo_columns(seattle_dataset: pd.DataFrame) -> None:
    """Test handling of missing geospatial columns."""
    data = seattle_dataset.drop(columns=["latitude", "longitude"])
    nn_index = NeuralNetworkIndex().create_index(
        data,
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["area", "baths", "beds"],
        preprocess_geo=True,
        num_models=1,
        num_epochs=1,
        min_pred_epoch=1,
    )
    assert nn_index is not None


@pytest.mark.usefixtures("seattle_dataset")
def test_nn_missing_date_column(seattle_dataset: pd.DataFrame) -> None:
    """Test handling of missing date column for temporal preprocessing."""
    data = seattle_dataset.copy().rename(columns={"sale_date": "transaction_date"})
    with pytest.raises(KeyError):
        _ = NeuralNetworkIndex().create_index(
            data,
            date="sale_date",  # this column no longer exists
            price="sale_price",
            trans_id="sale_id",
            prop_id="pinx",
            dep_var="price",
            ind_var=["area", "baths", "beds"],
            preprocess_time=True,
            num_models=1,
            num_epochs=1,
            min_pred_epoch=1,
        )


@pytest.mark.usefixtures("seattle_dataset")
def test_nn_model_inputs_to_features(seattle_dataset: pd.DataFrame) -> None:
    """Test conversion of model inputs to features."""
    nn_index = NeuralNetworkIndex().create_index(
        seattle_dataset,
        date="sale_date",
        price="sale_price",
        trans_id="sale_id",
        prop_id="pinx",
        dep_var="price",
        ind_var=["area", "baths", "beds"],
        num_models=1,
        num_epochs=1,
        min_pred_epoch=1,
    )

    # Create a mock data pipeline with preprocessors.
    data_pipeline = MagicMock()
    data_pipeline.feature_preprocessors_ = [
        GeospatialPreprocessor(
            resolutions=[6],
            latitude_col="latitude",
            longitude_col="longitude",
        ),
    ]

    # Check feature conversion.
    features = nn_index.model._model_inputs_to_features(data_pipeline, ["latitude", "longitude"])
    assert isinstance(features, list)
    assert len(features) > 0



import pandas as pd
import pytest

from hpipy.period_table import PeriodTable
from hpipy.trans_data import HedonicTransactionData, RepeatTransactionData


@pytest.mark.usefixtures("seattle_dataset")
def test_rt_create_trans(seattle_dataset: pd.DataFrame) -> None:
    """Test repeat transaction transaction data creation."""
    sales_df = PeriodTable(seattle_dataset).create_period_table(
        date="sale_date",
        periodicity="monthly",
    )

    repeat_trans_data = RepeatTransactionData(sales_df).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
    )
    assert len(repeat_trans_data.trans_df) == 5102

    # Min date with move.
    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        min_date="2012-03-21",
    )
    assert len(repeat_trans_data.trans_df) == 5102

    # Min date with adj.
    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        min_date="2012-03-21",
        adj_type="clip",
    )
    assert len(repeat_trans_data.trans_df) == 2827

    # Max with move.
    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        max_date="2015-03-21",
    )
    assert len(repeat_trans_data.trans_df) == 5102

    # Max with clip.
    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        max_date="2014-03-21",
        adj_type="clip",
    )
    assert len(repeat_trans_data.trans_df) == 1148

    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        seq_only=True,
    )
    assert len(repeat_trans_data.trans_df) == 4823

    repeat_trans_data = RepeatTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        seq_only=True,
        min_period_dist=12,
    )
    assert len(repeat_trans_data.trans_df) == 3795

    with pytest.raises(ValueError):
        RepeatTransactionData(seattle_dataset).create_transactions(
            prop_id="pinx",
            trans_id="sale_id",
            price="sale_price",
            date="sale_date",
            periodicity="mocnthly",
        )

    with pytest.raises(TypeError):
        RepeatTransactionData(seattle_dataset).create_transactions(
            prop_id="pinx",
            trans_id="sale_id",
            price="sale_price",
            date="sale_price",
            periodicity="monthly",
        )

    with pytest.raises(TypeError):
        RepeatTransactionData(seattle_dataset).create_transactions(
            date="sale_date",
            periodicity="monthly",
        )


@pytest.mark.usefixtures("seattle_dataset")
def test_hed_create_trans(seattle_dataset: pd.DataFrame) -> None:
    """Test hedonic transaction data creation."""
    sales_df = PeriodTable(seattle_dataset).create_period_table(
        date="sale_date",
        periodicity="monthly",
    )

    hedonic_trans_data = HedonicTransactionData(sales_df).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
    )
    assert len(hedonic_trans_data.trans_df) == len(sales_df.trans_df)

    # Min date with move.
    hedonic_trans_data = HedonicTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        min_date="2012-03-21",
    )
    assert len(hedonic_trans_data.trans_df) == len(sales_df.trans_df)

    # Min date with adj.
    hedonic_trans_data = HedonicTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        min_date="2012-03-21",
        adj_type="clip",
    )
    assert len(hedonic_trans_data.trans_df) == 34097

    # Max with move.
    hedonic_trans_data = HedonicTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        max_date="2015-03-21",
    )
    assert len(hedonic_trans_data.trans_df) == 43313

    # Max with clip.
    hedonic_trans_data = HedonicTransactionData(seattle_dataset).create_transactions(
        prop_id="pinx",
        trans_id="sale_id",
        price="sale_price",
        date="sale_date",
        periodicity="monthly",
        max_date="2014-03-21",
        adj_type="clip",
    )
    assert len(hedonic_trans_data.trans_df) == 21669



import os
import sys

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../.."))

project = "hpiPy"
author = "Reid Johnson"

try:
    import importlib.metadata

    __version__ = importlib.metadata.version("hpipy")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"

version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinxext.directives",
    "sphinxext_altair.altairplot",
]

templates_path = ["_templates"]
autosummary_generate = True
exclude_patterns = []

# HTML theme settings.
html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

# Output file base name for HTML help builder.
html_title = "hpiPy"
html_short_title = "hpiPy"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/hpipy-logo.svg"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

html_theme_options = {
    "navbar_start": ["navbar-logo", "navbar-project"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "primary_sidebar_end": [],
    "logo": {"image_dark": "_static/hpipy-logo.svg"},
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/reidjohnson/hpipy",
            "icon": "fab fa-github fa-lg",
            "type": "fontawesome",
        },
    ],
    "show_toc_level": 1,
}

html_context = {"default_mode": "light"}

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    "index": [],
    "releases/changes": [],
    "**": ["sidebar-nav-bs"],
}

# Autodoc settings.
autodoc_typehints = "description"
autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "special-members": "__init__",
    "undoc-members": True,
    "exclude-members": "__weakref__",
    "show-inheritance": True,
    "inherited-members": True,
}

# Autosummary settings.
autosummary_generate = True

# Intersphinx configuration.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pandas": ("http://pandas.pydata.org/docs/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# MyST parser settings.
myst_enable_extensions = [
    "colon_fence",
    "deflist",
]

# Copybutton settings.
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True



import uuid

from docutils import nodes
from sphinx.util.docutils import SphinxDirective


class InvisibleAltairPlot(SphinxDirective):
    has_content = True

    def run(self):
        import altair as alt

        code = "\n".join(self.content)
        ns = {}
        exec(code, ns)

        chart = ns.get("chart")
        if not isinstance(chart, alt.TopLevelMixin):
            msg = "Expected a variable named 'chart' with an Altair chart object."
            raise ValueError(msg)

        chart_id = f"vega-spec-{uuid.uuid4().hex}"
        spec = chart.to_json(indent=None)

        html = f"""
        <div class="altair-chart" id="container-{chart_id}">
            <script type="application/json" id="{chart_id}">
                {spec}
            </script>
            <script>
                (function() {{
                    const el = document.getElementById("{chart_id}");
                    if (el) {{
                        const spec = JSON.parse(el.textContent);
                        vegaEmbed(el.parentElement, spec, {{
                            actions: {{
                                export: true,
                                source: true,
                                editor: true,
                                compiled: false
                            }}
                        }}).catch(console.error);
                    }}
                }})();
            </script>
        </div>
        """

        return [nodes.raw("", html, format="html")]


def setup(app) -> None:
    app.add_directive("invisible-altair-plot", InvisibleAltairPlot)



"""Model utilities."""

import torch


def get_device(gpu_id: int | None = None) -> torch.device:
    """Get PyTorch device from GPU ID.

    Args:
        gpu_id (int | None, optional): GPU ID.
            Defaults to None.

    Returns:
        torch.device: Device based on GPU ID.

    """
    device: str = "cuda:0" if gpu_id is None else (f"cuda:{gpu_id}")
    return torch.device(device if torch.cuda.is_available() else "cpu")


def mixup(
    X: torch.Tensor,
    y: torch.Tensor,
    alpha: float = 1,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Mixup function for use as a data augmentation method.

    Implements mixup as described in [1], where new training samples are
    created by taking a convex combination of pairwise features and labels.

    Args:
        X (torch.Tensor): Input tensor representing training features.
        y (torch.Tensor): Input tensor representing training labels.
        alpha (float, optional): Beta distribution parameter.
            Defaults to 1.

    Returns:
        The mixed-up inputs, pairs of responses, and lambda.

    References:
        [1] Zhang, Hongyi, Moustapha Cisse, Yann N. Dauphin, and David
            Lopez-Paz. "mixup: Beyond Empirical Risk Minimization."
            International Conference on Learning Representations. ICLR, 2018.

    """
    indices = torch.randperm(y.size(0))
    beta = torch.distributions.Beta(alpha, alpha)
    # lam = beta.sample().to(get_device())
    lam = beta.sample(y.reshape(-1, 1).size()).to(get_device())
    X = lam * X + (1 - lam) * X[indices]
    y_a, y_b = y, y[indices]
    return X, y_a, y_b, lam


def quantile_loss(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    quantile: torch.Tensor,
) -> torch.Tensor:
    """Quantile regression or "tilted" loss.

    Args:
        y_true (torch.Tensor): True response values.
        y_pred (torch.Tensor): Predicted response values.
        quantile (torch.Tensor): Quantile to estimate.

    Returns:
        Quantile loss.

    """
    error = y_pred - y_true
    mask = (error.ge(0).float() - quantile).detach()
    return mask * error



import numpy as np
import pytest

from hpipy.utils.stineman_interpolation import (
    calculate_stineman_interpolant,
    fill_missing,
    interpolate_stineman,
    parabola_slope,
    stineman_slope,
)


def test_stineman_slope_basic() -> None:
    """Test basic functionality of stineman_slope."""
    x = np.array([1.0, 2.0, 3.0])
    y = np.array([1.0, 4.0, 9.0])  # y = x^2
    slopes = stineman_slope(x, y)
    assert len(slopes) == len(x)
    assert np.all(np.isfinite(slopes))
    # For y = x^2, slope should be approximately 2x.
    np.testing.assert_allclose(slopes, [2.0, 4.0, 6.0], rtol=0.3)


def test_stineman_slope_two_points() -> None:
    """Test stineman_slope with only two points."""
    x = np.array([1.0, 2.0])
    y = np.array([1.0, 2.0])  # linear function
    slopes = stineman_slope(x, y)
    assert len(slopes) == 2
    np.testing.assert_allclose(slopes, [1.0, 1.0])


def test_stineman_slope_with_scaling() -> None:
    """Test stineman_slope with scaling option."""
    x = np.array([1.0, 2.0, 3.0]) * 1000  # large x values
    y = np.array([1.0, 4.0, 9.0]) * 0.001  # small y values
    slopes = stineman_slope(x, y, scale=True)
    assert len(slopes) == len(x)
    assert np.all(np.isfinite(slopes))


def test_parabola_slope_two_points() -> None:
    """Test parabola_slope with only two points."""
    x = np.array([1.0, 2.0])
    y = np.array([1.0, 2.0])
    slopes = parabola_slope(x, y)
    assert len(slopes) == 2
    np.testing.assert_allclose(slopes, [1.0, 1.0])


def test_calculate_stineman_interpolant_basic() -> None:
    """Test basic functionality of calculate_stineman_interpolant."""
    x = np.array([1.0, 2.0, 3.0])
    y = np.array([1.0, 4.0, 9.0])  # y = x^2
    xout = np.array([1.5, 2.5])
    yout = calculate_stineman_interpolant(x, y, xout)
    assert len(yout) == len(xout)
    assert np.all(np.isfinite(yout))
    # Output should be close to x^2.
    np.testing.assert_allclose(yout, xout**2, rtol=0.2)


def test_calculate_stineman_interpolant_with_slopes() -> None:
    """Test calculate_stineman_interpolant with provided slopes."""
    x = np.array([1.0, 2.0, 3.0])
    y = np.array([1.0, 4.0, 9.0])
    yp = np.array([2.0, 4.0, 6.0])  # exact slopes for y = x^2
    xout = np.array([1.5, 2.5])
    yout = calculate_stineman_interpolant(x, y, xout, yp=yp)
    assert len(yout) == len(xout)
    assert np.all(np.isfinite(yout))


def test_calculate_stineman_interpolant_errors() -> None:
    """Test error handling in calculate_stineman_interpolant."""
    x = np.array([1.0, 2.0, 3.0])
    y = np.array([1.0, 4.0, 9.0])
    xout = np.array([1.5, 2.5])

    # Test non-increasing x.
    with pytest.raises(ValueError):
        calculate_stineman_interpolant(np.array([3.0, 2.0, 1.0]), y, xout)

    # Test mismatched lengths.
    with pytest.raises(ValueError):
        calculate_stineman_interpolant(x, y[:-1], xout)

    # Test NaN values.
    with pytest.raises(ValueError):
        calculate_stineman_interpolant(x, np.array([1.0, np.nan, 9.0]), xout)


def test_fill_missing_locf() -> None:
    """Test Last Observation Carried Forward."""
    x = np.array([1.0, np.nan, np.nan, 4.0, np.nan])
    filled = fill_missing(x, option="locf")
    expected = np.array([1.0, 1.0, 1.0, 4.0, 4.0])
    np.testing.assert_array_equal(filled, expected)


def test_fill_missing_nocb() -> None:
    """Test Next Observation Carried Backward."""
    x = np.array([np.nan, 2.0, np.nan, 4.0, np.nan])
    filled = fill_missing(x, option="nocb")
    expected = np.array([2.0, 2.0, 4.0, 4.0, np.nan])
    np.testing.assert_array_equal(filled, expected)


def test_interpolate_stineman_basic() -> None:
    """Test basic functionality of interpolate_stineman."""
    x = np.array([1.0, np.nan, 3.0, np.nan, 5.0])
    interpolated = interpolate_stineman(x)
    assert len(interpolated) == len(x)
    assert not np.any(np.isnan(interpolated))
    # Values at known points should remain unchanged.
    np.testing.assert_allclose(interpolated[::2], [1.0, 3.0, 5.0])


def test_interpolate_stineman_no_nans() -> None:
    """Test interpolate_stineman with no missing values."""
    x = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    interpolated = interpolate_stineman(x)
    np.testing.assert_array_equal(interpolated, x)



"""Period table. Represents time periods for house price indices."""

import datetime
import logging
from typing import Any

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
import pandas as pd


def check_date(
    date: str | datetime.date | datetime.datetime | pd.Series | None,
    name: str,
) -> None | datetime.datetime:
    """Check and validate the provided date object.

    Checks that the date is either of type datetime or convertible to
    datetime. Provides error handling for incorrect date inputs.

    Args:
        date (str | datetime.date | datetime.datetime | pd.Series | None):
            Date object that needs validation.
        name (str): Name for the 'date' argument being checked.

    Returns:
        None | datetime.datetime: Date converted to a datetime.datetime
            object. If the input date is None, the function returns None.

    Raises:
        ValueError: If date cannot be converted to a datetime object.

    """
    # If None, return None.
    if date is None:
        return None

    # If a number, give an error (stops from converting numbers to dates).
    # if not np.issubdtype(date.dtype, np.datetime64):
    #     raise TypeError("Date must be a datetime type.")

    # If any form of date/time, convert to datetime.
    if isinstance(date, datetime.date | datetime.datetime):
        date = pd.to_datetime(date)
    else:
        # Try to convert dates given as characters, such as '2000-01-01'.
        date = pd.to_datetime(date)
        if np.any(date is pd.NaT):
            msg = f"{name} argument must be a datetime type."
            raise ValueError(msg)

    return date


class PeriodTable:
    """Period table class.

    Organizes transaction data into defined time periods.

    """

    trans_df: pd.DataFrame
    period_table: pd.DataFrame

    def __init__(
        self,
        trans_df: pd.DataFrame,
        period_table: pd.DataFrame | None = None,
    ) -> None:
        """Initialize period table.

        Args:
            trans_df (pd.DataFrame): Transaction data.
            period_table (pd.DataFrame | None): Period table.

        Raises:
            TypeError: If 'trans_df' is not or does not inherit from a DataFrame.

        """
        # Check that `trans_df` is a DataFrame.
        if not isinstance(trans_df, pd.DataFrame):
            msg = "'trans_df' must be a DataFrame (or inherit from one)."
            raise TypeError(msg)

        self.trans_df = trans_df

        if period_table is not None:
            self.period_table = period_table

    @staticmethod
    def _check_date(
        trans_df: pd.DataFrame,
        date: str,
        min_date: str | None,
        max_date: str | None,
        adj_type: str | None,
    ) -> tuple[pd.DataFrame, datetime.datetime, datetime.datetime]:
        """Check date fields and arguments.

        Args:
            trans_df (pd.DataFrame): Transaction data.
            date (str): Date field.
            min_date (str | None): Minimum date.
            max_date (str | None): Maximum date.
            adj_type (str | None): Adjustment type.

        Returns:
            tuple[pd.DataFrame, datetime.datetime, datetime.datetime]:
                Transformed transaction data, minimum date, and maximum date.

        """
        if not np.issubdtype(trans_df[date].dtype, np.datetime64):
            msg = "Date must be a datetime type."
            raise TypeError(msg)

        # Check date fields.
        trans_df.loc[:, date] = check_date(trans_df[date], "date")
        min_datetime = check_date(min_date, "min_date")
        max_datetime = check_date(max_date, "max_date")

        # Set minimum date.
        if min_datetime is None:
            min_datetime = trans_df[date].min()
        elif min_datetime > trans_df[date].min():
            if adj_type == "move":
                logging.info(
                    "Supplied 'min_date' is greater than minimum of transactions. Adjusting.",
                )
                min_datetime = trans_df[date].min()
            elif adj_type == "clip":
                logging.info(
                    "Supplied 'min_date' date is greater than minimum of transactions. "
                    "Clipping transactions.",
                )
                trans_df = trans_df[trans_df[date] >= min_datetime].copy()

        # Set maximum date.
        if max_datetime is None:
            max_datetime = trans_df[date].max()
        elif max_datetime < trans_df[date].max():
            if adj_type == "move":
                logging.info(
                    "Supplied 'max_date' is less than maximum of transactions. Adjusting.",
                )
                max_datetime = trans_df[date].max()
            elif adj_type == "clip":
                logging.info(
                    "Supplied 'max_date' is less than maximum of transactions. "
                    "Clipping transactions.",
                )
                trans_df = trans_df[trans_df[date] <= max_datetime]

        # Set standardized date field.
        trans_df.loc[:, "trans_date"] = trans_df[date]

        return trans_df, min_datetime, max_datetime  # type: ignore

    @staticmethod
    def _check_periodicity(periodicity: str | None) -> str:
        """Check periodicity argument.

        Args:
            periodicity (str | None): Periodicity.

        Returns:
            str: Periodicity.

        """
        # Check for periodicity.
        if periodicity is None:
            logging.warning("No 'periodicity' supplied, defaulting to 'annual'.")
            periodicity = "annual"

        periodicity_options = [
            "weekly",
            "monthly",
            "quarterly",
            "annual",
            "yearly",
            "w",
            "m",
            "q",
            "a",
            "y",
            "equalfreq",
            "ef",
            "equalsample",
            "es",
        ]
        periodicity = periodicity.lower()
        if periodicity not in periodicity_options:
            msg = (
                "'periodicity' must be one of: 'weekly', 'monthly', 'quarterly', "
                "'annual', 'equalfreq' or 'equalsample'."
            )
            raise ValueError(msg)
        if periodicity in ["yearly", "y", "a"]:
            periodicity = "annual"
        elif periodicity == "q":
            periodicity = "quarterly"
        elif periodicity == "m":
            periodicity = "monthly"
        elif periodicity == "w":
            periodicity = "weekly"
        elif periodicity == "ef":
            periodicity = "equalfreq"
        elif periodicity == "es":
            periodicity = "equalsample"

        return periodicity

    @staticmethod
    def _create_periods(
        trans_df: pd.DataFrame,
        periodicity: str,
        nbr_periods: int | None = None,
        freq: int | None = None,
        start: str | None = None,
        first_date: str | None = None,
        last_date: str | None = None,
    ) -> pd.DataFrame:
        """Create dataframe of periods of specified periodicity.

        Args:
            trans_df (pd.DataFrame): Transaction data.
            periodicity (str): Periodicity.
            nbr_periods (int | None): Number of periods.
            freq (int | None): Frequency.
            start (str | None): Start.
            first_date (str | None): First date.
            last_date (str | None): Last date.

        Returns:
            pd.DataFrame: Period table.

        """
        if periodicity == "annual":
            date_range = pd.date_range(
                trans_df["trans_date"].min().to_period("Y").to_timestamp(),
                trans_df["trans_date"].max().to_period("Y").to_timestamp(),
                freq="YS",
            )
            start_date = list(date_range)
            end_date = list(date_range.to_period("Y").to_timestamp(how="end"))
            period = list(range(1, len(start_date) + 1))
            name = [f"{x.strftime('%Y')}" for x in start_date]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}
        elif periodicity == "quarterly":
            date_range = pd.date_range(
                trans_df["trans_date"].min().to_period("Q").to_timestamp(),
                trans_df["trans_date"].max().to_period("Q").to_timestamp(),
                freq="QS",
            )
            start_date = list(date_range)
            end_date = list(date_range.to_period("Q").to_timestamp(how="end"))
            period = list(range(1, len(start_date) + 1))
            name = [
                # f"{x_start.strftime('%Y-%q')}"
                f"{x_start.to_period('Q').strftime('%Y-%q')}"  # bug workaround
                for x_start in start_date
            ]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}
        elif periodicity == "monthly":
            date_range = pd.date_range(
                trans_df["trans_date"].min().to_period("M").to_timestamp(),
                trans_df["trans_date"].max().to_period("M").to_timestamp(),
                freq="MS",
            )
            start_date = list(date_range)
            end_date = list(date_range.to_period("M").to_timestamp(how="end"))
            period = list(range(1, len(start_date) + 1))
            name = [x.strftime("%Y-%b") for x in start_date]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}
        elif periodicity == "weekly":
            date_range = pd.date_range(
                trans_df["trans_date"].min().to_period("W").to_timestamp(),
                trans_df["trans_date"].max().to_period("W").to_timestamp(),
                freq="W-MON",
            )
            start_date = date_range.map(lambda x: x - pd.DateOffset(days=1)).tolist()
            end_date = (
                date_range.to_series().apply(lambda dt: dt + pd.DateOffset(weekday=5)).tolist()
            )
            period = list(range(1, len(start_date) + 1))
            name = [
                f"week: {x_start.strftime('%Y-%m-%d')} to {x_end.strftime('%Y-%m-%d')}"
                for x_start, x_end in zip(start_date, end_date, strict=False)
            ]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}
        elif periodicity == "equalfreq":
            if first_date is None:
                first = trans_df["trans_date"].min()
            else:
                first = pd.to_datetime(first_date)
                if first > trans_df["trans_date"].min():
                    logging.info("'first_date' is within bounds of data.")
            if last_date is None:
                last = trans_df["trans_date"].max()
            else:
                last = pd.to_datetime(last_date)
                if last < trans_df["trans_date"].max():
                    logging.info("'last_date' is within bounds of data.")

            if freq is None:
                freq = 30
                logging.warning(
                    "You did not supply a frequency ('freq = '). "
                    "Running at the default of 30 days.",
                )

            if start is None:
                start = "first"
                logging.warning(
                    "You did not specify when you wanted to start counting ('start = "
                    "['first' | 'last']). Defaulting to starting from the first sale.",
                )

            if start == "last":
                date_range = pd.date_range(
                    last.to_period("D").to_timestamp(),
                    first.to_period("D").to_timestamp(),
                    freq=f"-{freq}D",
                )
                date_range = date_range[::-1]
            else:
                date_range = pd.date_range(
                    first.to_period("D").to_timestamp(),
                    last.to_period("D").to_timestamp(),
                    freq=f"{freq}D",
                )
            start_date = list(date_range)
            end_date = list(pd.Series(date_range) + datetime.timedelta(days=freq))

            if start == "last" and start_date[0] > min(first, trans_df["trans_date"].min()):
                start_date[0] = min(first, trans_df["trans_date"].min())
            if end_date[-1] > last:
                i = -1
                while end_date[i - 1] > last:
                    i -= 1
                end_date[i - 1] = last
                start_date = start_date[:i]
                end_date = end_date[:i]
            period = list(range(1, len(start_date) + 1))
            name = [
                f"equalfreq ({freq}): {x_start.strftime('%Y-%m-%d')} to "
                f"{x_end.strftime('%Y-%m-%d')}"
                for x_start, x_end in zip(start_date, end_date, strict=False)
            ]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}
        elif periodicity == "equalsample":
            if nbr_periods is None:
                raise ValueError
            # Set quantiles.
            period_qtls = [(x / nbr_periods) for x in range(nbr_periods + 1)]
            date_qtls = trans_df["trans_date"].astype("int64").quantile(period_qtls[:-1]).values
            date_qtls[0] = date_qtls[0] - 1
            start_date = pd.to_datetime(
                pd.Series(
                    pd.to_datetime(date_qtls, origin="1970-01-01") + pd.to_timedelta(1, unit="s"),
                ).dt.strftime("%Y-%m-%d"),
            ).tolist()
            end_date = (
                pd.to_datetime(
                    pd.concat(
                        [
                            pd.Series(pd.to_datetime(date_qtls[1:], origin="1970-01-01")),
                            pd.Series(pd.to_datetime(trans_df["trans_date"].max())),
                        ],
                        ignore_index=True,
                    ).dt.strftime("%Y-%m-%d"),
                )
                - datetime.timedelta(days=1)
            ).to_list()
            period = list(range(1, nbr_periods + 1))
            name = [f"period {x}" for x in list(range(1, nbr_periods + 1))]
            data = {"start_date": start_date, "end_date": end_date, "period": period, "name": name}

        return pd.DataFrame(data)

    def create_period_table(
        self,
        date: str,
        periodicity: str | None = None,
        nbr_periods: int | None = None,
        freq: int | None = None,
        start: str | None = None,
        min_date: str | None = None,
        max_date: str | None = None,
        adj_type: str | None = "move",
        **kwargs: Any,
    ) -> Self:
        """Create a period table from a transaction dataframe.

        Args:
            date (str): Date field.
            periodicity (str | None, optional): Periodicity of the table.
                Defaults to None.
            nbr_periods (int | None, optional): Number of periods (only
                used if `periodicity` is "equalsample").
                Defaults to None.
            freq (int | None, optional): Frequency in days (only used if
                `periodicity` is "equalfreq").
                Defaults to None.
            start (str | None, optional): Starting position (only used if
                `periodicity` is "equalfreq"). One of "first" or "last".
                Defaults to None.
            min_date (str | None, optional): Minimum date.
                Defaults to None.
            max_date (str | None, optional): Maximum dare.
                Defaults to None.
            adj_type (str | None, optional): Adjustment type.
                Defaults to "move".

        Returns:
            Self: Period table.

        """
        trans_df = self.trans_df

        trans_df, min_datetime, max_datetime = self._check_date(
            trans_df,
            date,
            min_date,
            max_date,
            adj_type,
        )
        periodicity = self._check_periodicity(periodicity)

        # Make period_table.
        period_table = self._create_periods(
            trans_df,
            periodicity,
            nbr_periods=nbr_periods,
            freq=freq,
            start=start,
            **kwargs,
        )

        # Add to `trans_df`.
        trans_df.loc[:, "trans_period"] = (
            pd.cut(
                trans_df["trans_date"],
                pd.concat(
                    [
                        period_table["start_date"],
                        period_table.loc[len(period_table) - 1 :, "end_date"]
                        + datetime.timedelta(days=1),
                    ],
                    ignore_index=True,
                ),
                right=False,
                labels=False,
            )
            + 1
        ).astype(float)

        # Check for missing periods.
        num_periods = trans_df["trans_period"].nunique()
        if num_periods < len(period_table):
            logging.info(
                f"Your choice of periodicity resulted in {len(period_table) - num_periods} "
                f"empty periods out of {len(period_table)} total periods.",
            )
            if (len(period_table) - num_periods) / len(period_table) > 0.3:
                logging.warning(
                    "You may wish to set a coarser periodicity "
                    "or set different start and end dates.",
                )

        self.trans_df = trans_df
        self.period_table = period_table

        self.min_date = min_datetime.strftime("%Y-%m-%d")
        self.max_date = max_datetime.strftime("%Y-%m-%d")

        return self



"""Custom model layers."""

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn


class MonotonicDense(nn.Linear):
    """Monotonic counterpart of the regular dense (linear) layer.

    Implements a monotonically constrained layer as described in [1].

    References:
        [1] Runje, Davor, and Sharath M. Shankaranarayana. "Constrained
            Monotonic Neural Networks." International Conference on Machine
            Learning. PMLR, 2023.

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        activation: nn.Module | None = None,
        monotonicity_indicator: int | list[int] | None = 1,
        activation_weights: list[int] | None = None,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the layer.

        Args:
            in_features (int): Size of each input sample.
            out_features (int): Size of each output sample.
            activation (Optional[nn.Module], optional): Activation function.
                Defaults to None.
            monotonicity_indicator (int | list[int] | None, optional):
                Vector to indicate which of the inputs are monotonically
                increasing, monotonically decreasing, or non-monotonic. Has
                value 1 for monotonically increasing, -1 for monotonically
                decreasing, and 0 for non-monotonic variables.
                Defaults to 1.
            activation_weights (list[int] | None = None, optional): Activation
                weights for convex, concave, and saturated units, respectively.
                Defaults to None. If None, defaults to [1, 1, 1].
            device (torch.device | None, optional): Device on which to
                initialize the layer. Defaults to None.
            dtype (torch.dtype | None, optional): Data type of the layer.
                Defaults to None.

        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__(in_features, out_features, **factory_kwargs, **kwargs)  # type: ignore

        self.activation = activation

        if monotonicity_indicator is not None:
            self.register_buffer(
                "monotonicity_indicator",
                torch.tensor(monotonicity_indicator, **factory_kwargs),  # type: ignore
            )
        else:
            self.monotonicity_indicator = None

        if activation_weights is None:
            activation_weights = [1, 1, 1]
        else:
            if len(activation_weights) != 3:
                msg = (
                    "`activation_weights` must have exactly three components, got "
                    f"{activation_weights}."
                )
                raise ValueError(msg)
            if (np.array(activation_weights) < 0).any():
                msg = (
                    "`activation_weights` values must be non-negative, got "
                    f"{activation_weights}."
                )
                raise ValueError(msg)
        self.activation_weights = torch.Tensor(activation_weights, device=device)
        self.activation_weights = self.activation_weights / self.activation_weights.sum(dim=0)

        self.n_convex = round(self.activation_weights[0].item() * self.out_features)
        self.n_concave = round(self.activation_weights[1].item() * self.out_features)
        self.n_saturated = self.out_features - self.n_convex - self.n_concave

    @contextmanager
    def _replace_kernel(self) -> Generator:
        """Replace kernel weights with signed values as indicated."""
        original_weight = self.weight

        if self.monotonicity_indicator is not None:
            weight_abs = torch.abs(self.weight)

            if self.monotonicity_indicator.dim == 0:
                if torch.all(self.monotonicity_indicator == 1):
                    self.weight.data = weight_abs
                elif torch.all(self.monotonicity_indicator == -1):
                    self.weight.data = -weight_abs
            else:
                w_ = torch.where(self.monotonicity_indicator == 1, weight_abs, self.weight)
                self.weight.data = torch.where(self.monotonicity_indicator == -1, -weight_abs, w_)
        yield

        self.weight = original_weight

    def _activation_selector(self, x: torch.Tensor) -> torch.Tensor:
        """Select activation sign."""
        return torch.cat(
            [
                torch.ones((x.shape[0], self.n_convex), dtype=torch.int, device=x.device),
                -torch.ones((x.shape[0], self.n_concave), dtype=torch.int, device=x.device),
                torch.zeros((x.shape[0], self.n_saturated), dtype=torch.int, device=x.device),
            ],
            dim=1,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            x (torch.Tensor): Input data.

        Returns:
            Output data.

        """
        with self._replace_kernel():
            y = super().forward(x)

        if self.activation is not None:
            activation_selector = self._activation_selector(x)

            c = self.activation(torch.ones_like(y))

            y1 = self.activation(y)
            y2 = -self.activation(-y)
            y3 = torch.where(y <= 0, self.activation(y + 1) - c, -self.activation(-(y - 1)) + c)

            y = torch.where(activation_selector == 1, y1, y)
            y = torch.where(activation_selector == -1, y2, y)
            y = torch.where(activation_selector == 0, y3, y)

        return y


class OrdinalEmbedding(nn.Module):
    """Ordinal embedding layer.

    Embeds an ordinal variable using a cumulative or "binary counting"
    encoding, where each ordinal value is encoded as an cumulative binary
    array. Null values are assumed to be the value zero; they are represented
    by an encoding consisting of only zeros and are handled separately by the
    embedding layer from non-null encodings.

    For example, the values [1, 3, 2, 0, 4] are encoded as:

        [[1, 0, 0, 0],
         [1, 1, 1, 0],
         [1, 1, 0, 0],
         [0, 0, 0, 0],
         [1, 1, 1, 1]]

    The embedding applies a single linear layer to all non-zero encodings and
    a separate set of parameters of the same output size to the zero encoding.
    The outputs are additively combined to produce the final embedding output.
    """

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        """Initialize the layer.

        Args:
            num_embeddings (int): Size of the dictionary of embeddings.
            embedding_dim (int): Size of each embedding vector.
            device (torch.device | None, optional): Device on which to
                initialize the layer. Defaults to None.
            dtype (torch.dtype | None, optional): Data type of the layer.
                Defaults to None.

        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.weight = nn.Parameter(
            torch.empty(
                (self.embedding_dim, self.num_embeddings),
                **factory_kwargs,  # type: ignore
            ),
        )
        self.bias: torch.Tensor | None
        self.register_parameter("bias", None)
        self.null_weight = nn.Parameter(
            torch.empty(self.embedding_dim, **factory_kwargs),  # type: ignore
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Reset the layer parameters."""
        std = 0.01  # small value given the additive effects of the encoding
        nn.init.normal_(self.weight, std=std)
        nn.init.normal_(self.null_weight, std=std)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            x (torch.Tensor): Input data.

        Returns:
            Output data.

        """
        x = x.reshape(-1, 1)
        nonnull_mask = x.any(dim=1)
        x = (x > torch.arange(self.num_embeddings, device=x.device)).float().to(x.device)
        x_out = torch.zeros(x.size(0), self.embedding_dim, device=x.device)
        x_out[nonnull_mask] += F.linear(x[nonnull_mask], self.weight, self.bias)
        x_out[~nonnull_mask] += self.null_weight
        return x_out

    def extra_repr(self) -> str:
        """Format display representation.

        Returns:
            String representation.

        """
        s = "{num_embeddings}, {embedding_dim}"
        return s.format(**self.__dict__)



"""Random forest HPI extensions."""

import logging
from typing import Any

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
import pandas as pd
from quantile_forest import RandomForestQuantileRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.inspection import partial_dependence

from hpipy.price_index import HedonicIndex
from hpipy.price_model import BaseHousePriceModel


class RandomForestIndex(HedonicIndex):
    """Random forest house price index."""

    @staticmethod
    def get_model() -> type[BaseHousePriceModel]:
        """Get a random forest house price model."""
        return RandomForestModel


class RandomForestModel(BaseHousePriceModel):
    """Random forest house price model."""

    def _create_model(
        self,
        X: pd.DataFrame | pd.Series,
        y: pd.Series,
        **kwargs: Any,
    ) -> RandomForestRegressor | RandomForestQuantileRegressor:
        """Create a random forest house price model.

        Args:
            X (pd.DataFrame | pd.Series): Independent variables.
            y (pd.Series): Dependent variable.
            **kwargs: Additional keyword arguments.

        Returns:
            RandomForestRegressor | RandomForestQuantileRegressor: Random forest model.

        """
        return self._model_with_coefficients(X, y, **kwargs)

    def _model_with_coefficients(
        self,
        X: pd.DataFrame | pd.Series,
        y: pd.Series,
        estimator: str,
        n_estimators: int,
        random_seed: int,
        **kwargs: Any,
    ) -> RandomForestRegressor | RandomForestQuantileRegressor:
        """Fit model and populate coefficients to produce index.

        Args:
            X (pd.DataFrame | pd.Series): Independent variables.
            y (pd.Series): Dependent variable.
            estimator (str): Estimator type.
            n_estimators (int): Number of estimators.
            random_seed (int): Random seed.
            **kwargs: Additional keyword arguments.

        Returns:
            RandomForestRegressor | RandomForestQuantileRegressor: Random forest model.

        """
        # Fit the model.
        model = RandomForestRegressor(n_estimators=n_estimators, random_state=random_seed).fit(
            X,
            y,
        )
        # model = RangerForestRegressor(
        #    n_estimators=n_estimators,
        #    categorical_features=["trans_period"],
        #    seed=random_seed,
        # ).fit(X, y)
        if estimator == "pdp":
            return self._model_pdp(model, X, y, random_seed=random_seed, **kwargs)
        raise ValueError

    def _model_pdp(
        self,
        model: RandomForestRegressor | RandomForestQuantileRegressor,
        X: pd.DataFrame | pd.Series,
        y: pd.Series,
        log_dep: bool = False,
        random_seed: int = 0,
        **kwargs: Any,
    ) -> RandomForestRegressor | RandomForestQuantileRegressor:
        """Generate explanation from partial dependence plot (PDP) values.

        Args:
            model (RandomForestRegressor | RandomForestQuantileRegressor): Random forest model.
            X (pd.DataFrame | pd.Series): Independent variables.
            y (pd.Series): Dependent variable.
            log_dep (bool, optional): Log transform the dependent variable.
                Defaults to False.
            random_seed (int): Random seed.
            **kwargs: Additional keyword arguments.

        Returns:
            RandomForestRegressor | RandomForestQuantileRegressor: Random forest model.

        """
        # Get simulation dataframe.
        sim_X, _ = self._create_sim_df(X, y, random_seed=random_seed, **kwargs)

        # col_indices = [idx for idx, x in enumerate(X.columns) if x.startswith("trans_period_")]
        # partial_dependencies = []
        # for col_idx in range(len(X.columns)):
        #     predictions = partial_dependence(
        #         model, sim_X, col_idx, categorical_features=col_indices
        #     )
        #     partial_dependencies.append(predictions["average"].squeeze()[-1])

        col_idx = list(X.columns).index("trans_period")
        predictions = partial_dependence(model, sim_X, col_idx, categorical_features=[col_idx])
        values = predictions["grid_values"][0]
        partial_dependencies = predictions["average"].squeeze()

        pdp_df = (
            pd.DataFrame({"yhat": np.nan}, index=np.arange(X["trans_period"].max()) + 1)
            .combine_first(pd.DataFrame({"yhat": partial_dependencies}, index=values))
            .interpolate(fill_value="ffill")
            .bfill()
        )

        base_yhat = pdp_df["yhat"].iloc[0]

        # Add coefficients.
        coefs = pdp_df["yhat"] - base_yhat if log_dep else pdp_df["yhat"] / base_yhat

        model.coef_ = coefs

        return model

    def _create_sim_df(
        self,
        X: pd.DataFrame | pd.Series,
        y: pd.Series,
        random_seed: int,
        sim_ids: int | None = None,
        sim_count: int | None = None,
        sim_per: float | None = None,
        **kwargs: Any,
    ) -> tuple[pd.DataFrame | pd.Series, pd.Series]:
        """Create simulation data.

        Args:
            X (pd.DataFrame | pd.Series): Independent variables.
            y (pd.Series): Dependent variable.
            random_seed (int): Random seed.
            sim_ids (int | None, optional): Simulation IDs.
            sim_count (int | None, optional): Simulation count.
            sim_per (float | None, optional): Simulation percentage.

        Returns:
            tuple[pd.DataFrame | pd.Series, pd.Series]: Simulation data.

        """
        # If no filters.
        if sim_ids is None and sim_count is None and sim_per is None:
            return X, y

        # If by sim id.
        if sim_ids is not None:
            return X.iloc[sim_ids], y[sim_ids]

        # If a sim percentage is provided.
        if sim_count is None:
            sim_count = int(np.floor(sim_per * len(X))) if sim_per is not None else len(X)

        # Take sample.
        np.random.seed(random_seed)
        indices = np.random.randint(len(X), size=sim_count)
        X = X.iloc[indices[:sim_count]]
        y = y[indices[:sim_count].tolist()]

        return X, y

    def fit(
        self,
        dep_var: str | None = None,
        ind_var: list[str] | None = None,
        estimator: str = "pdp",
        log_dep: bool = True,
        n_estimators: int = 100,
        random_seed: int = 0,
        **kwargs: Any,
    ) -> Self:
        """Fit the random forest model and generate index coefficients.

        Args:
            dep_var (str | None, optional): Dependent variable.
                Defaults to None.
            ind_var (list[str] | None, optional): Independent variable(s).
                Defaults to None.
            estimator (str, optional): Estimator type.
                Defaults to "pdp".
            log_dep (bool, optional): Log transform the dependent variable.
                Defaults to True.
            n_estimators (int, optional): Number of estimators.
                Defaults to 100.
            random_seed (int, optional): Random seed to use.
                Defaults to 0.

        Returns:
            Self: Fitted model.

        """
        hpi_df = self.hpi_df.copy()

        if dep_var is None or ind_var is None:
            msg = "'dep_var' and 'ind_var' must be supplied."
            raise ValueError(msg)

        # oh_enc = OneHotEncoder(drop="first")
        # oh_enc.fit(hpi_df[["trans_period"]])
        # X_cats = pd.DataFrame(
        #     oh_enc.transform(hpi_df[["trans_period"]]).toarray(),
        #     columns=oh_enc.get_feature_names_out(["trans_period"]),
        # )
        # X = pd.concat([hpi_df[ind_var].reset_index(drop=True), X_cats], axis=1)

        for var in ind_var:
            if hpi_df[var].dtype == "object":
                hpi_df[var] = hpi_df[var].astype("category")
                hpi_df[var] = hpi_df[var].cat.codes

        X = hpi_df[[*ind_var, "trans_period"]].reset_index(drop=True)
        y = np.log(hpi_df[dep_var]) if log_dep else hpi_df[dep_var]

        # Extract base period mean price.
        base_price = hpi_df["price"][hpi_df["trans_period"] == hpi_df["trans_period"].min()].mean()

        # Check for legal estimator type.
        if estimator not in ["pdp"]:
            logging.warning(
                "Provided estimator type is not supported. Allowed estimators are: "
                "'pdp'. Defaulting to 'pdp'.",
            )
            estimator = "pdp"

        # Check log dep vs data.
        if (log_dep and np.any(hpi_df["price"] <= 0)) or (hpi_df["price"].isnull().sum() > 0):
            msg = "Your 'price' field includes invalid values."
            raise ValueError(msg)

        model = self._create_model(
            X=X,
            y=y,
            estimator=estimator,
            n_estimators=n_estimators,
            log_dep=log_dep,
            random_seed=random_seed,
            **kwargs,
        )

        # Check for successful model estimation.
        if not isinstance(model, (RandomForestRegressor | RandomForestQuantileRegressor)):
            msg = "Model estimator was unsuccessful."
            raise ValueError(msg)

        # Period names.
        # p_names = [len(ind_var) + idx for idx in range(len(oh_enc.categories_[0]) - 1)]
        # periods = list(range(1, len(oh_enc.categories_[0]) + 1))
        periods = list(np.arange(hpi_df["trans_period"].max()) + 1)

        # Coefficients.
        # coefs = np.hstack([[0], model.coef_[p_names]])
        coefs = model.coef_

        model_df = pd.DataFrame({"time": periods, "coefficient": coefs})

        self.coefficients = model_df
        self.model_obj = model
        self.periods = self.period_table
        self.base_price = base_price
        self.params = {
            "estimator": estimator,
            "log_dep": log_dep,
            "ind_var": ind_var,
            "dep_var": dep_var,
            "n_estimators": n_estimators,
            "random_seed": random_seed,
        }
        self.X = X
        self.y = y

        return self



import importlib.resources as pkg_resources

import pandas as pd


def load_ex_sales() -> pd.DataFrame:
    """Load example home sales dataset.

    Returns:
        pandas.DataFrame: A DataFrame with house sales.

    """
    with pkg_resources.files("hpipy.datasets.data").joinpath("ex_sales.csv").open("r") as f:
        return pd.read_csv(f, parse_dates=["sale_date"])


def load_seattle_sales() -> pd.DataFrame:
    """Load Seattle home sales dataset.

    Returns:
        pandas.DataFrame: A DataFrame with house sales.

    """
    with pkg_resources.files("hpipy.datasets.data").joinpath("seattle_sales.csv").open("r") as f:
        return pd.read_csv(f, parse_dates=["sale_date"])
